
from typing import Tuple, List

import datetime
from sqlalchemy import distinct
from sqlalchemy.sql import func
import pandas as pd
import numpy as np
from ...util.singleton import Singleton
from ...constant import FOFStrategyTypeDic
from ..wrapper.mysql import BasicDatabaseConnector
from ..view.basic_models import *
from .raw import RawDataApi


class BasicDataApi(metaclass=Singleton):
    def get_trading_day_list(self, start_date='', end_date=''):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        TradingDayList
                    )
                if start_date:
                    query = query.filter(
                        TradingDayList.datetime >= start_date,
                    )
                if end_date:
                    query = query.filter(
                        TradingDayList.datetime <= end_date,
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {TradingDayList.__tablename__}')

    def get_last_week_last_trade_dt(self):
        today = datetime.datetime.today().date()
        start_date = today - datetime.timedelta(days=15)
        df = self.get_trading_day_list(start_date=start_date)
        df.loc[:,'weekday'] = df.datetime.map(lambda x : x.weekday())
        td = pd.to_datetime(df.datetime)
        td = [_.week for _ in td]
        df.loc[:,'week_num'] = td
        last_week_num = df.week_num.values[-1]
        last_week_dt = df[df.week_num != last_week_num].datetime.values[-1]
        return last_week_dt

    def is_today_trading_date(self):
        today = datetime.datetime.now().date()
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        TradingDayList
                    ).filter(
                        TradingDayList.datetime >= today,
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                con = today in tag_df.datetime.tolist()
                return con

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {TradingDayList.__tablename__}')

    def get_last_trading_day(self, dt=datetime.datetime.today()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query_result = db_session.query(
                    func.max(TradingDayList.datetime)
                ).filter(
                    TradingDayList.datetime < dt.date(),
                ).one_or_none()
                if query_result:
                    return query_result[0]
                else:
                    return None

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {TradingDayList.__tablename__}')

    def get_sector_info(self):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        SectorInfo
                    )
                ret_df = pd.read_sql(query.statement, query.session.bind)
                return ret_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {SectorInfo.__tablename__}')

    def get_sector_index_funds(self, index_id):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        SectorFunds
                ).filter(
                    SectorFunds.index_id == index_id,
                )
                ret_df = pd.read_sql(query.statement, query.session.bind)
                return ret_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {SectorFunds.__tablename__}')

    def get_sector_index_info(self):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        SectorFunds
                )
                ret_df = pd.read_sql(query.statement, query.session.bind)
                return ret_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {SectorFunds.__tablename__}')

    def get_sector_main_index_id(self, sector_id):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    SectorInfo.main_index_id
                ).filter(
                    SectorInfo.sector_id == sector_id
                )
                ret_df = pd.read_sql(query.statement, query.session.bind)
                return ret_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {SectorInfo.__tablename__}')

    def get_sector_indices(self, sector_id):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    distinct(SectorFunds.index_id).label("index_id")
                ).filter(
                    SectorFunds.sector_id == sector_id
                )
                ret_df = pd.read_sql(query.statement, query.session.bind)
                return ret_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {SectorFunds.__tablename__}')

    def get_fund_info(self, fund_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundInfo
                    )
                if fund_list:
                    query = query.filter(
                        FundInfo.fund_id.in_(fund_list),
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundInfo.__tablename__}')

    def delete_fund_info(self, fund_list: Tuple[str]):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundInfo
                ).filter(FundInfo.fund_id.in_(fund_list))
                query.delete(synchronize_session=False)
                db_session.commit()
            except Exception as e:
                print(f'Failed to delete data <err_msg> {e} from {FundInfo.__tablename__}')

    def get_fund_info_by_key_words(self, key_words: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundInfo
                )
                for key_word in key_words:
                    query = query.filter(
                         FundInfo.desc_name.like(f'%{key_word}%')
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundInfo.__tablename__}')
                return None

    def get_fund_benchmark(self, fund_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundBenchmark
                )
                if fund_list:
                    query = query.filter(
                        FundBenchmark.em_id.in_(fund_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundBenchmark.__tablename__}')

    def delete_fund_benchmark(self, fund_list: Tuple[str]):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundBenchmark
                ).filter(FundBenchmark.em_id.in_(fund_list))
                query.delete(synchronize_session=False)
                db_session.commit()
            except Exception as e:
                print(f'Failed to delete data <err_msg> {e} from {FundBenchmark.__tablename__}')

    def get_index_info(self, index_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        IndexInfo
                    )
                if index_list:
                    query = query.filter(
                        IndexInfo.index_id.in_(index_list)
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {IndexInfo.__tablename__}')

    def get_index_info_by_key_words(self, key_words:Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        IndexInfo
                    )
                for key_word in key_words:
                    query = query.filter(
                        IndexInfo.desc_name.like(f'%{key_word}%')
                    )
                df = pd.read_sql(query.statement, query.session.bind)
                return df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {IndexInfo.__tablename__}')

    def get_index_info_by_em_id(self, em_id_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        IndexInfo
                    )
                if em_id_list:
                    query = query.filter(
                        IndexInfo.em_id.in_(em_id_list)
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {IndexInfo.__tablename__}')

    def get_index_component(self, index_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    IndexComponent
                )
                if index_list:
                    query = query.filter(
                        IndexComponent.index_id.in_(index_list)
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {IndexComponent.__tablename__}')
                return None

    def get_stock_info(self):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        StockInfo
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {StockInfo.__tablename__}')

    def get_fund_nav(self, fund_list=None, dt=None):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundNav
                )
                if fund_list:
                    query = query.filter(
                        FundNav.fund_id.in_(fund_list),
                    )
                if dt:
                    query = query.filter(
                        FundNav.datetime == dt,
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundNav.__tablename__}')

    def delete_fund_nav(self, start_date, end_date, fund_list: Tuple[str]):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundNav
                ).filter(
                    FundNav.datetime >= start_date,
                    FundNav.datetime <= end_date,
                )
                if fund_list:
                    query = query.filter(FundNav.fund_id.in_(fund_list))
                query.delete(synchronize_session=False)
                db_session.commit()
                return True
            except Exception as e:
                print(f'Failed to delete data <err_msg> {e} from {FundNav.__tablename__}')
                return None

    def get_fund_nav_with_date(self, start_date='', end_date='', fund_list: Tuple[str] = (), columns: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                if columns:
                    query = db_session.query(
                        FundNav.fund_id,
                        FundNav.datetime,
                    ).add_columns(*columns)
                else:
                    query = db_session.query(
                        FundNav,
                    )
                if fund_list:
                    query = query.filter(FundNav.fund_id.in_(fund_list))
                if start_date:
                    query = query.filter(
                        FundNav.datetime >= start_date,
                    )
                if end_date:
                    query = query.filter(
                        FundNav.datetime <= end_date,
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundNav.__tablename__}')

    def get_stock_price(self, stock_list, begin_date: str = '', end_date: str = ''):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        StockPrice
                ).filter(
                    # We could query all fund_ids at one time
                    StockPrice.stock_id.in_(stock_list),
                )
                if begin_date:
                    query = query.filter(StockPrice.datetime >= begin_date)
                if end_date:
                    query = query.filter(StockPrice.datetime <= end_date)

                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {StockPrice.__tablename__}')

    def get_fund_ret(self, fund_list):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundRet
                ).filter(
                        # We could query all fund_ids at one time
                        FundRet.fund_id.in_(fund_list),
                    )

                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundRet.__tablename__}')

    def get_index_price(self, index_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        IndexPrice
                )
                if index_list:
                    query = query.filter(
                        # We could query all fund_ids at one time
                        IndexPrice.index_id.in_(index_list),
                    )

                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {IndexPrice.__tablename__}')

    def get_index_price_dt(self, start_date: str = '', end_date: str = '', index_list: Tuple[str] = (), columns: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                if columns:
                    query = db_session.query(
                        IndexPrice.index_id,
                        IndexPrice.datetime,
                    ).add_columns(*columns)
                else:
                    query = db_session.query(
                        IndexPrice
                    )
                if index_list:
                    query = query.filter(
                        IndexPrice.index_id.in_(index_list),
                    )
                if start_date:
                    query = query.filter(
                        IndexPrice.datetime >= start_date,
                    )
                if end_date:
                    query = query.filter(
                        IndexPrice.datetime <= end_date,
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {IndexPrice.__tablename__}')

    def delete_index_price(self, index_id_list, start_date, end_date):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                db_session.query(
                    IndexPrice
                ).filter(
                    IndexPrice.index_id.in_(index_id_list),
                    IndexPrice.datetime >= start_date,
                    IndexPrice.datetime <= end_date,
                ).delete(synchronize_session=False)
                db_session.commit()
                return True
            except Exception as e:
                print(f'Failed to delete data <err_msg> {e} from {IndexPrice.__tablename__}')
                return None

    def get_fund_nav_with_date_range(self, fund_list, start_date, end_date):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundNav
                ).filter(
                        FundNav.fund_id.in_(fund_list),
                        FundNav.datetime >= start_date,
                        FundNav.datetime <= end_date,
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundNav.__tablename__}')

    def get_fund_list(self):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundInfo.fund_id
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundInfo.__tablename__}')

    def get_fund_fee(self):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundInfo.fund_id,
                        FundInfo.manage_fee,
                        FundInfo.trustee_fee,
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundInfo.__tablename__}')

    def get_fund_asset(self):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundInfo.fund_id,
                        FundInfo.asset_type,
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundInfo.__tablename__}')

    def get_fund_id_mapping(self):
        fund_id_mapping = {}
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query_results = db_session.query(
                        FundInfo.fund_id,
                        FundInfo.order_book_id,
                        FundInfo.start_date,
                        FundInfo.end_date
                    ).all()

                for fund_id, order_book_id, start_date, end_date in query_results:
                    if order_book_id not in fund_id_mapping:
                        fund_id_mapping[order_book_id] = []
                    fund_id_mapping[order_book_id].append(
                        {'fund_id': fund_id, 'start_date':start_date, 'end_date': end_date})

            except Exception as e:
                print(f'Failed to get fund id mapping <err_msg> {e} from {FundInfo.__tablename__}')
                return None

        return fund_id_mapping

    def get_fund_size(self):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundSize,
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundSize.__tablename__}')

    def get_fund_open_info(self):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundOpenInfo,
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundOpenInfo.__tablename__}')


    def get_fund_status(self):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundStatusLatest,
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundStatusLatest.__tablename__}')

    def get_inside_market_funds(self):
        fund_status_latest = self.get_fund_status()
        in_market_fund_list = fund_status_latest[(~fund_status_latest['trade_status'].isnull()) & (fund_status_latest.trade_status != '终止上市')].fund_id.tolist()
        fund_info = self.get_fund_info()
        return fund_info[fund_info.fund_id.isin(in_market_fund_list)]

    def get_outside_market_funds(self):
        fund_status_latest = self.get_fund_status()
        in_market_fund_list = fund_status_latest[(~fund_status_latest['trade_status'].isnull()) & (fund_status_latest.trade_status != '终止上市')].fund_id.tolist()
        fund_info = self.get_fund_info()
        return fund_info[~fund_info.fund_id.isin(in_market_fund_list)]

    def get_fund_conv_stats(self):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundConvStats,
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e}')

    @staticmethod
    def get_history_fund_size(fund_id: str = '', start_date=None, end_date=None):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    Fund_size_and_hold_rate,
                )
                if fund_id:
                    query = query.filter(
                        Fund_size_and_hold_rate.fund_id == fund_id
                    )
                if start_date is not None:
                    query = query.filter(
                        Fund_size_and_hold_rate.datetime >= start_date,
                    )
                if end_date is not None:
                    query = query.filter(
                        Fund_size_and_hold_rate.datetime <= end_date,
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {Fund_size_and_hold_rate.__tablename__}')
                return pd.DataFrame([])

    def delete_fund_size_hold_rate(self, date_to_delete: datetime.date, fund_list: List[str]):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                db_session.query(
                    Fund_size_and_hold_rate
                ).filter(
                    Fund_size_and_hold_rate.fund_id.in_(fund_list),
                    Fund_size_and_hold_rate.datetime == date_to_delete,
                ).delete(synchronize_session=False)
                db_session.commit()
                return True
            except Exception as e:
                print(f'Failed to delete data <err_msg> {e} from {Fund_size_and_hold_rate.__tablename__}')
                return False

    @staticmethod
    def get_fund_size_range(start_date='', end_date='', fund_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    Fund_size_and_hold_rate.fund_id,
                    Fund_size_and_hold_rate.datetime,
                    Fund_size_and_hold_rate.size,
                )
                if fund_list:
                    query = query.filter(
                        Fund_size_and_hold_rate.fund_id.in_(fund_list),
                    )
                if start_date:
                    query = query.filter(
                        # 按报告日多选取3个月，保证初值可以fill出来
                        Fund_size_and_hold_rate.datetime >= (pd.to_datetime(start_date) - datetime.timedelta(days=100)).strftime('%Y%m%d'),
                    )
                if end_date:
                    query = query.filter(
                        Fund_size_and_hold_rate.datetime <= end_date,
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {Fund_size_and_hold_rate.__tablename__}')
                return pd.DataFrame([])

    @staticmethod
    def get_fund_company_hold(start_date='', end_date='', fund_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    Fund_size_and_hold_rate.fund_id,
                    Fund_size_and_hold_rate.datetime,
                    Fund_size_and_hold_rate.institution_holds,
                )
                if fund_list:
                    query = query.filter(
                        Fund_size_and_hold_rate.fund_id.in_(fund_list),
                    )
                if start_date:
                    query = query.filter(
                        # 按报告日多选取3个月，保证初值可以fill出来
                        Fund_size_and_hold_rate.datetime >= (pd.to_datetime(start_date) - datetime.timedelta(days=100)).strftime('%Y%m%d'),
                    )
                if end_date:
                    query = query.filter(
                        Fund_size_and_hold_rate.datetime <= end_date,
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {Fund_size_and_hold_rate.__tablename__}')
                return pd.DataFrame([])

    @staticmethod
    def get_fund_persional_hold(start_date='', end_date='', fund_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    Fund_size_and_hold_rate.fund_id,
                    Fund_size_and_hold_rate.datetime,
                    Fund_size_and_hold_rate.personal_holds,
                )
                if fund_list:
                    query = query.filter(
                        Fund_size_and_hold_rate.fund_id.in_(fund_list),
                    )
                if start_date:
                    query = query.filter(
                        # 按报告日多选取3个月，保证初值可以fill出来
                        Fund_size_and_hold_rate.datetime >= (pd.to_datetime(start_date) - datetime.timedelta(days=100)).strftime('%Y%m%d'),
                    )
                if end_date:
                    query = query.filter(
                        Fund_size_and_hold_rate.datetime <= end_date,
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {Fund_size_and_hold_rate.__tablename__}')
                return pd.DataFrame([])

    @staticmethod
    def get_fund_size_all_data():
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    Fund_size_and_hold_rate
                )
                df = pd.read_sql(query.statement, query.session.bind)
                return df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {Fund_size_and_hold_rate.__tablename__}')
                return pd.DataFrame([])

    @staticmethod
    def get_history_fund_rating(fund_id):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundRate,
                ).filter(
                    FundRate.fund_id == fund_id
                )
                df = pd.read_sql(query.statement, query.session.bind)
                return df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundRate.__tablename__}')
                return pd.DataFrame([])

    @staticmethod
    def get_fund_hold_asset_by_id(fund_id: str = '', start_date=None, end_date=None):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundHoldAsset,
                )
                if fund_id:
                    query = query.filter(
                        FundHoldAsset.fund_id == fund_id,
                    )
                if start_date is not None:
                    query = query.filter(
                        FundHoldAsset.datetime >= start_date,
                    )
                if end_date is not None:
                    query = query.filter(
                        FundHoldAsset.datetime <= end_date,
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundHoldAsset.__tablename__}')
                return pd.DataFrame([])


    def get_fund_hold_asset_by_id_list(self, fund_id):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundHoldAsset,
                )
                if fund_id:
                    query = query.filter(
                        FundHoldAsset.fund_id.in_(fund_id),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundHoldAsset.__tablename__}')
                return pd.DataFrame([])


    def delete_fund_hold_asset(self, date_to_delete: datetime.date, fund_list: List[str]):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                db_session.query(
                    FundHoldAsset
                ).filter(
                    FundHoldAsset.fund_id.in_(fund_list),
                    FundHoldAsset.datetime == date_to_delete,
                ).delete(synchronize_session=False)
                db_session.commit()
                return True
            except Exception as e:
                print(f'Failed to delete data <err_msg> {e} from {FundHoldAsset.__tablename__}')
                return False

    @staticmethod
    def get_fund_hold_industry_by_id(fund_id):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundHoldIndustry,
                ).filter(
                    FundHoldIndustry.fund_id == fund_id,
                )
                df = pd.read_sql(query.statement, query.session.bind)
                return df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundHoldIndustry.__tablename__}')
                return pd.DataFrame([])

    @staticmethod
    def get_fund_hold_stock_by_id(fund_id: str = ''):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundHoldStock,
                )
                if fund_id:
                    query = query.filter(
                        FundHoldStock.fund_id == fund_id,
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundHoldStock.__tablename__}')
                return pd.DataFrame([])

    def get_fund_hold_stock_by_id_list(self, fund_id):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundHoldStock,
                )
                if fund_id:
                    query = query.filter(
                        FundHoldStock.fund_id.in_(fund_id),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundHoldStock.__tablename__}')
                return pd.DataFrame([])


    @staticmethod
    def get_fund_hold_bond_by_id(fund_id):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundHoldBond,
                ).filter(
                    FundHoldBond.fund_id == fund_id,
                )
                df = pd.read_sql(query.statement, query.session.bind)
                return df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundHoldBond.__tablename__}')
                return pd.DataFrame([])

    def get_fund_hold_bond_by_id_list(self, fund_id):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundHoldBond,
                ).filter(
                    FundHoldBond.fund_id.in_(fund_id),
                )
                df = pd.read_sql(query.statement, query.session.bind)
                return df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundHoldBond.__tablename__}')
                return pd.DataFrame([])


    def get_style_analysis_data(self, start_date: str, end_date: str, stock_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    StyleAnalysisStockFactor
                )
                if stock_list:
                    query = query.filter(
                        StyleAnalysisStockFactor.stock_id.in_(stock_list)
                    )
                query = query.filter(
                    StyleAnalysisStockFactor.datetime >= start_date,
                    StyleAnalysisStockFactor.datetime <= end_date,
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {StyleAnalysisStockFactor.__tablename__}')
                return

    def get_style_analysis_time_range(self) -> pd.DataFrame:
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    func.max(StyleAnalysisStockFactor.datetime).label('end_date'),
                    func.min(StyleAnalysisStockFactor.datetime).label('start_date'),
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get time range data <err_msg> {e} from {StyleAnalysisStockFactor.__tablename__}')
                return

    def get_barra_cne5_risk_factor(self, start_date: str, end_date: str, stock_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    BarraCNE5RiskFactor
                )
                if stock_list:
                    query = query.filter(
                        BarraCNE5RiskFactor.stock_id.in_(stock_list)
                    )
                query = query.filter(
                    BarraCNE5RiskFactor.datetime >= start_date,
                    BarraCNE5RiskFactor.datetime <= end_date,
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {BarraCNE5RiskFactor.__tablename__}')
                return

    def get_fund_hold_asset(self, dt) -> pd.DataFrame:
       with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundHoldAsset
                        ).filter(
                            FundHoldAsset.datetime == dt
                        )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get time range data <err_msg> {e} from {FundHoldAsset.__tablename__}')
                return

    def get_fund_hold_industry(self, dt) -> pd.DataFrame:
       with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundHoldIndustry
                        ).filter(
                            FundHoldIndustry.datetime == dt
                        )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get time range data <err_msg> {e} from {FundHoldIndustry.__tablename__}')
                return

    def delete_fund_hold_industry(self, date_to_delete: datetime.date, fund_list: List[str]):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                db_session.query(
                    FundHoldIndustry
                ).filter(
                    FundHoldIndustry.fund_id.in_(fund_list),
                    FundHoldIndustry.datetime == date_to_delete,
                ).delete(synchronize_session=False)
                db_session.commit()
                return True
            except Exception as e:
                print(f'Failed to delete data <err_msg> {e} from {FundHoldIndustry.__tablename__}')
                return False

    def get_fund_hold_stock(self, dt=None) -> pd.DataFrame:
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundHoldStock
                )
                if dt:
                    query = query.filter(
                        FundHoldStock.datetime == dt
                    )
                else:
                    query = query.filter(
                        FundHoldStock.datetime > datetime.datetime.now() - datetime.timedelta(days=180)
                    )

                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get time range data <err_msg> {e} from {FundHoldStock.__tablename__}')
                return

    def get_fund_hold_stock_latest(self, dt=None, fund_list =[]) -> pd.DataFrame:
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                if not dt:
                    dt = db_session.query(
                        func.max(FundHoldStock.datetime)
                    ).filter(
                    FundHoldStock.fund_id.in_(fund_list),
                ).one_or_none()
                    if dt:
                        dt = dt[0]

                query = db_session.query(
                    FundHoldStock
                ).filter(
                    FundHoldStock.datetime == dt,
                    FundHoldStock.fund_id.in_(fund_list),
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get time range data <err_msg> {e} from {FundHoldStock.__tablename__}')
                return

    def get_fund_hold_bond_latest(self, dt=None, fund_list =[]) -> pd.DataFrame:
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                if not dt:
                    dt = db_session.query(
                        func.max(FundHoldBond.datetime)
                    ).filter(
                    FundHoldBond.fund_id.in_(fund_list),
                ).one_or_none()
                    if dt:
                        dt = dt[0]
                query = db_session.query(
                    FundHoldBond
                ).filter(
                    FundHoldBond.datetime == dt,
                    FundHoldBond.fund_id.in_(fund_list),
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get time range data <err_msg> {e} from {FundHoldBond.__tablename__}')
                return

    def get_fund_hold_asset_latest(self, dt=None, fund_list = []) -> pd.DataFrame:
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                if not dt:
                    dt = db_session.query(
                        func.max(FundHoldAsset.datetime)
                    ).filter(
                    FundHoldAsset.fund_id.in_(fund_list),
                ).one_or_none()
                    if dt:
                        dt = dt[0]

                query = db_session.query(
                        FundHoldAsset
                        ).filter(
                            FundHoldAsset.datetime == dt,
                            FundHoldAsset.fund_id.in_(fund_list),
                        )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get time range data <err_msg> {e} from {FundHoldAsset.__tablename__}')
                return

    def delete_fund_hold_stock(self, date_to_delete: datetime.date, fund_list: List[str]):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                db_session.query(
                    FundHoldStock
                ).filter(
                    FundHoldStock.fund_id.in_(fund_list),
                    FundHoldStock.datetime == date_to_delete,
                ).delete(synchronize_session=False)
                db_session.commit()
                return True
            except Exception as e:
                print(f'Failed to delete data <err_msg> {e} from {FundHoldStock.__tablename__}')
                return False

    def get_fund_hold_bond(self, dt: str = '') -> pd.DataFrame:
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundHoldBond
                )
                if dt:
                    query = query.filter(
                        FundHoldBond.datetime == dt
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get time range data <err_msg> {e} from {FundHoldBond.__tablename__}')
                return

    def get_fund_ipo_stats(self, dt: str = '') -> pd.DataFrame:
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundIPOStats
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get time range data <err_msg> {e} from {FundIPOStats.__tablename__}')
                return

    def delete_fund_hold_bond(self, date_to_delete: datetime.date, fund_list: List[str]):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                db_session.query(
                    FundHoldBond
                ).filter(
                    FundHoldBond.fund_id.in_(fund_list),
                    FundHoldBond.datetime == date_to_delete,
                ).delete(synchronize_session=False)
                db_session.commit()
                return True
            except Exception as e:
                print(f'Failed to delete data <err_msg> {e} from {FundHoldBond.__tablename__}')
                return False

    def get_fund_stock_portfolio(self, dt: str = ''):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundStockPortfolio
                )
                if dt:
                    query = query.filter(
                        FundStockPortfolio.datetime == dt
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get time range data <err_msg> {e} from {FundStockPortfolio.__tablename__}')
                return

    def get_fund_stock_concept(self, tag_group_id_list: List[str]):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    StockTag
                )
                if tag_group_id_list:
                    query = query.filter(
                        StockTag.tag_group_id.in_(tag_group_id_list)
                    )
                return pd.read_sql(query.statement, query.session.bind).drop(columns=['_update_time'])
            except Exception as e:
                print(f'Failed to get_fund_stock_concept <err_msg> {e} from {StockTag.__tablename__}')
                return

    def delete_fund_rate(self, date_to_delete: datetime.date, fund_list: List[str]):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                db_session.query(
                    FundRate
                ).filter(
                    FundRate.fund_id.in_(fund_list),
                    FundRate.datetime == date_to_delete,
                ).delete(synchronize_session=False)
                db_session.commit()
                return True
            except Exception as e:
                print(f'Failed to delete data <err_msg> {e} from {FundRate.__tablename__}')

    @staticmethod
    def get_model_data(model: Base, *params):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        *(getattr(model, k) for k in params if hasattr(model, k))
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {model.__tablename__}')

    @classmethod
    def get_fund_info_data(cls, *params):
        return cls.get_model_data(FundInfo, *params)


    def get_hedge_fund_info(self, fund_id_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    HedgeFundInfo,
                )
                if fund_id_list:
                    query = query.filter(
                        HedgeFundInfo.fund_id.in_(fund_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_hedge_fund_info <err_msg> {e} from {HedgeFundInfo.__tablename__}')

    def get_hedge_fund_alias(self, manager_id: str, fund_id_alias: str):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    HedgeFundAlias,
                ).filter(
                    HedgeFundAlias.fund_id_alias == fund_id_alias,
                    HedgeFundAlias.manager_id == manager_id,
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_hedge_fund_alias <err_msg> {e} from {HedgeFundAlias.__tablename__}')

    def get_hedge_fund_nav(self, fund_id_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    HedgeFundNAV,
                )
                if fund_id_list:
                    query = query.filter(
                        HedgeFundNAV.fund_id.in_(fund_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_hedge_fund_nav <err_msg> {e} from {HedgeFundNAV.__tablename__}')


    def delete_hedge_fund_nav(self, fund_id_to_delete: str, *, date_list: Tuple[datetime.date] = (), start_date: str = '', end_date: str = '') -> bool:
        with BasicDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    HedgeFundNAV
                ).filter(
                    HedgeFundNAV.fund_id == fund_id_to_delete,
                )
                if date_list:
                    query = query.filter(
                        HedgeFundNAV.datetime.in_(date_list),
                    )
                if start_date:
                    query = query.filter(
                        HedgeFundNAV.datetime >= start_date,
                    )
                if end_date:
                    query = query.filter(
                        HedgeFundNAV.datetime <= end_date,
                    )
                query.delete(synchronize_session=False)
                session.commit()
                return True
            except Exception as e:
                print(f'failed to delete data <err_msg> {e} from {HedgeFundNAV.__tablename__}')
                return False


    def get_fof_info(self, manager_id: str, fof_id_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FOFInfo,
                ).filter(
                    FOFInfo.manager_id == manager_id,
                )
                if fof_id_list:
                    query = query.filter(
                        FOFInfo.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_info <err_msg> {e} from {FOFInfo.__tablename__}')

    def get_fof_info_by_key_words(self, manager_id: str, key_words:Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FOFInfo,
                ).filter(
                    FOFInfo.manager_id == manager_id,
                )
                for key_word in key_words:
                    query = query.filter(
                        FOFInfo.fof_name.like(f'%{key_word}%')
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_info <err_msg> {e} from {FOFInfo.__tablename__}')
    
    def get_fof_info_strategy_exsit(self, manager_id:str='1', fof_id_list:Tuple[str]=(), has_nav:bool=True):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FOFInfo,
                ).filter(
                    FOFInfo.manager_id == manager_id,
                    FOFInfo.has_nav == has_nav,
                    FOFInfo.strategy_type.isnot(None),
                )
                if fof_id_list:
                    query = query.filter(
                        FOFInfo.fof_id.in_(fof_id_list),
                    )
                df = pd.read_sql(query.statement, query.session.bind)
                df.strategy_type = df.strategy_type.map(lambda x:FOFStrategyTypeDic[x] )
                return df
            except Exception as e:
                print(f'failed to get_fof_info_nav_exsited <err_msg> {e} from {FOFInfo.__tablename__}')

    def get_fof_info_ignore_other_info(self, manager_id:str='1', fof_id_list:Tuple[str]=()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FOFInfo,
                ).filter(
                    FOFInfo.manager_id == manager_id,
                )
                if fof_id_list:
                    query = query.filter(
                        FOFInfo.fof_id.in_(fof_id_list),
                    )
                df = pd.read_sql(query.statement, query.session.bind)
                df.strategy_type = df.strategy_type.map(lambda x:FOFStrategyTypeDic[x] )
                return df
            except Exception as e:
                print(f'failed to get_fof_info <err_msg> {e} from get_fof_info_ignore_other_info')

    def get_fof_asset_allocation(self, manager_id, fof_id_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FOFAssetAllocation,
                ).filter(
                    FOFAssetAllocation.manager_id == manager_id,
                )
                if fof_id_list:
                    query = query.filter(
                        FOFAssetAllocation.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_asset_allocation <err_msg> {e} from {FOFAssetAllocation.__tablename__}')


    def get_fof_scale_alteration(self, fof_id_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FOFScaleAlteration,
                )
                if fof_id_list:
                    query = query.filter(
                        FOFScaleAlteration.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_scale_alteration <err_msg> {e} from {FOFScaleAlteration.__tablename__}')


    def get_fof_manually(self, fof_id_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FOFManually,
                )
                if fof_id_list:
                    query = query.filter(
                        FOFManually.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_manually <err_msg> {e} from {FOFManually.__tablename__}')

    def get_fof_other_record(self, fof_id_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FOFOtherRecord,
                )
                if fof_id_list:
                    query = query.filter(
                        FOFOtherRecord.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_manually <err_msg> {e} from {FOFOtherRecord.__tablename__}')

    def get_fof_account_statement(self, manager_id: str, fof_id_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FOFAccountStatement,
                ).filter(
                    FOFAccountStatement.manager_id == manager_id,
                )
                if fof_id_list:
                    query = query.filter(
                        FOFAccountStatement.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_account_statement <err_msg> {e} from {FOFAccountStatement.__tablename__}')

    def delete_fof_account_statement(self, date_to_delete: datetime.date, fof_id_list: List[str]) -> bool:
        with BasicDatabaseConnector().managed_session() as session:
            try:
                session.query(
                    FOFAccountStatement
                ).filter(
                    FOFAccountStatement.fof_id.in_(fof_id_list),
                    FOFAccountStatement.datetime == date_to_delete,
                ).delete(synchronize_session=False)
                session.commit()
                return True
            except Exception as e:
                print(f'failed to delete data <err_msg> {e} from {FOFAccountStatement.__tablename__}')
                return False

    def get_fof_incidental_statement(self, manager_id, fof_id_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FOFIncidentalStatement,
                ).filter(
                    FOFIncidentalStatement.manager_id == manager_id,
                )
                if fof_id_list:
                    query = query.filter(
                        FOFIncidentalStatement.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_incidental_statement <err_msg> {e} from {FOFIncidentalStatement.__tablename__}')

    def get_fof_investor_position(self, manager_id: str, fof_id_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FOFInvestorPosition,
                ).filter(
                    FOFInvestorPosition.manager_id == manager_id,
                )
                if fof_id_list:
                    query = query.filter(
                        FOFInvestorPosition.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_investor_position <err_msg> {e} from {FOFInvestorPosition.__tablename__}')

    def delete_fof_investor_position(self, manager_id: str, fof_id_to_delete: datetime.date, investor_id_list: List[str]) -> bool:
        with BasicDatabaseConnector().managed_session() as session:
            try:
                session.query(
                    FOFInvestorPosition
                ).filter(
                    FOFInvestorPosition.fof_id == fof_id_to_delete,
                    FOFInvestorPosition.manager_id == manager_id,
                    FOFInvestorPosition.investor_id.in_(investor_id_list),
                ).delete(synchronize_session=False)
                session.commit()
                return True
            except Exception as e:
                print(f'failed to delete data <err_msg> {e} from {FOFInvestorPosition.__tablename__}')
                return False

    def get_fof_investor_position_summary(self, manager_id: str, investor_id_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FOFInvestorPositionSummary,
                ).filter(
                    FOFInvestorPositionSummary.manager_id == manager_id,
                )
                if investor_id_list:
                    query = query.filter(
                        FOFInvestorPositionSummary.investor_id.in_(investor_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_investor_position_summary <err_msg> {e} from {FOFInvestorPositionSummary.__tablename__}')

    def delete_fof_investor_position_summary(self, manager_id: str, investor_id: str, date_list: List[str]) -> bool:
        with BasicDatabaseConnector().managed_session() as session:
            try:
                session.query(
                    FOFInvestorPositionSummary
                ).filter(
                    FOFInvestorPositionSummary.manager_id == manager_id,
                    FOFInvestorPositionSummary.investor_id == investor_id,
                    FOFInvestorPositionSummary.datetime.in_(date_list),
                ).delete(synchronize_session=False)
                session.commit()
                return True
            except Exception as e:
                print(f'failed to delete data <err_msg> {e} from {FOFInvestorPositionSummary.__tablename__}')

    def get_fof_estimate_fee(self, fof_id_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FOFEstimateFee,
                )
                if fof_id_list:
                    query = query.filter(
                        FOFEstimateFee.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_estimate_fee <err_msg> {e} from {FOFEstimateFee.__tablename__}')

    def delete_fof_estimate_fee(self, date_to_delete: datetime.date, fof_id_list: List[str]) -> bool:
        with BasicDatabaseConnector().managed_session() as session:
            try:
                session.query(
                    FOFEstimateFee
                ).filter(
                    FOFEstimateFee.fof_id.in_(fof_id_list),
                    FOFEstimateFee.date == date_to_delete,
                ).delete(synchronize_session=False)
                session.commit()
                return True
            except Exception as e:
                print(f'failed to delete data <err_msg> {e} from {FOFEstimateFee.__tablename__}')
                return False

    def get_fof_estimate_interest(self, manager_id: str, fof_id_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FOFEstimateInterest,
                ).filter(
                    FOFEstimateInterest.manager_id == manager_id,
                )
                if fof_id_list:
                    query = query.filter(
                        FOFEstimateInterest.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_estimate_interest <err_msg> {e} from {FOFEstimateInterest.__tablename__}')

    def delete_fof_estimate_interest(self, date_to_delete: datetime.date, fof_id_list: List[str]) -> bool:
        with BasicDatabaseConnector().managed_session() as session:
            try:
                session.query(
                    FOFEstimateInterest
                ).filter(
                    FOFEstimateInterest.fof_id.in_(fof_id_list),
                    FOFEstimateInterest.date == date_to_delete,
                ).delete(synchronize_session=False)
                session.commit()
                return True
            except Exception as e:
                print(f'failed to delete data <err_msg> {e} from {FOFEstimateInterest.__tablename__}')
                return False

    def get_fof_transit_money(self, fof_id_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FOFTransitMoney,
                )
                if fof_id_list:
                    query = query.filter(
                        FOFTransitMoney.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_transit_money <err_msg> {e} from {FOFTransitMoney.__tablename__}')

    def delete_fof_transit_money(self, date_to_delete: datetime.date, fof_id_list: List[str]) -> bool:
        with BasicDatabaseConnector().managed_session() as session:
            try:
                session.query(
                    FOFTransitMoney
                ).filter(
                    FOFTransitMoney.fof_id.in_(fof_id_list),
                    FOFTransitMoney.confirmed_datetime == date_to_delete,
                ).delete(synchronize_session=False)
                session.commit()
                return True
            except Exception as e:
                print(f'failed to delete data <err_msg> {e} from {FOFTransitMoney.__tablename__}')
                return False

    def get_fof_assert_correct(self, manager_id: str, fof_id: str):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FOFAssertCorrect,
                ).filter(
                    FOFAssertCorrect.manager_id == manager_id,
                    FOFAssertCorrect.fof_id == fof_id,
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_assert_correct <err_msg> {e} from {FOFAssertCorrect.__tablename__}')

    def get_email_parser_code(self, p_id_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    EMailParserCode,
                )
                if p_id_list:
                    query = query.filter(
                        EMailParserCode.id.in_(p_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_email_parser_code <err_msg> {e} from {EMailParserCode.__tablename__}')

    def get_email_parser_exec(self, manager_id: str):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    EMailParserExec,
                ).filter(
                    EMailParserExec.manager_id.in_([manager_id, None]),
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_email_parser_manager <err_msg> {e} from {EMailParserExec.__tablename__}')

    def get_hedge_fund_email_raw(self, manager_id: str, fof_id: str, fund_id_list: List[str]):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    HedgeFundEmailRaw,
                ).filter(
                    HedgeFundEmailRaw.manager_id == manager_id,
                    HedgeFundEmailRaw.fof_id == fof_id,
                    HedgeFundEmailRaw.fund_id.in_(fund_id_list),
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_hedge_fund_email_raw <err_msg> {e} from {HedgeFundEmailRaw.__tablename__}')

    def get_asset_info(self, asset_id_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    AssetInfo,
                )
                if asset_id_list:
                    query = query.filter(
                        AssetInfo.asset_id.in_(asset_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_transit_money <err_msg> {e} from {AssetInfo.__tablename__}')

    def get_asset_info_real(self, real_id_list: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    AssetInfo,
                )
                if real_id_list:
                    query = query.filter(
                        AssetInfo.real_id.in_(real_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_transit_money <err_msg> {e} from {AssetInfo.__tablename__}')

    def get_asset_info_by_real_id_and_asset_type(self, real_id_list: Tuple[str] = (), asset_type:Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    AssetInfo,
                )
                if asset_type:
                    query = query.filter(
                        AssetInfo.asset_type.in_(asset_type),
                        AssetInfo.real_id.in_(real_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_transit_money <err_msg> {e} from {AssetInfo.__tablename__}')

    def get_asset_info_by_type(self, asset_type: Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    AssetInfo,
                )
                if asset_type:
                    query = query.filter(
                        AssetInfo.asset_type.in_(asset_type),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_transit_money <err_msg> {e} from {AssetInfo.__tablename__}')

    def get_asset_info_by_type_by_key_words(self,asset_type: Tuple[str] = (), key_words:Tuple[str] = ()):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    AssetInfo,
                )
                if asset_type:
                    query = query.filter(
                        AssetInfo.asset_type.in_(asset_type),
                    )
                for key_word in key_words:
                    query = query.filter(
                        AssetInfo.real_name.like(f'%{key_word}%')
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_transit_money <err_msg> {e} from {AssetInfo.__tablename__}')

    def get_asset_details(self):
        try:
            df = self.get_asset_info()
            asset_list = df.asset_type.unique().tolist()
            result = []
            for asset_id in asset_list:
                dic = {
                    'title':asset_id,
                    'items':[],
                }
                _df = df[df.asset_type == asset_id].drop_duplicates(subset=['asset_name'])
                for r in _df.itertuples():
                    _dic = {'id':r.asset_id,'name':r.asset_name}
                    dic['items'].append(_dic)
                result.append(dic)
            return result
        except Exception as e:
            print(f'failed to get_fof_transit_money <err_msg> {e} from {AssetInfo.__tablename__}')

    def get_product_details(self):
        try:
            '''
            df = self.get_asset_info()
            df1 = df[(df['asset_type'] == '典型产品')]
            asset_list = df1.asset_name.unique().tolist()
            result = {}
            for asset_id in asset_list:
                result[asset_id] = []
                _df = df1[df1.asset_name == asset_id]
                for r in _df.itertuples():
                    result[asset_id].append([r.real_id, r.real_name])
            '''
            fund_info = self.get_fof_info_strategy_exsit()
            # 客户专属
            special_list = ['景林资产','石锋资产','林园投资','东方港湾','万方资产']
            fund_info_1 = fund_info[~fund_info.admin.isin(special_list)]
            fund_info_2 = fund_info[fund_info.admin.isin(special_list)]
            asset_list = fund_info.strategy_type.unique().tolist()
            result = {}
            for asset_id in asset_list:
                result[asset_id] = []
                _df = fund_info_1[fund_info_1.strategy_type == asset_id]
                for r in _df.itertuples():
                    result[asset_id].append([r.fof_id, r.fof_name])
            df = self.get_asset_info()
            df2 = df[df['asset_name'].isin(['商品','大盘股','中证500','创业板指','中证800','中证1000','国债'])]
            asset_list = df2.asset_name.unique().tolist()
            dic = {'商品CFCI':'商品期货指数'}
            benchmark_tag = '*业绩基准'
            result[benchmark_tag] = []
            for r in df2.itertuples():
                result[benchmark_tag].append([r.real_id, dic.get(r.real_name,r.real_name)])
            for agency_i in special_list:
                _agency_i = f'#{agency_i}'
                result[_agency_i] = []
                _df = fund_info_2[fund_info_2.admin == agency_i]
                for r in _df.itertuples():
                    result[_agency_i].append([r.fof_id, r.fof_name])
            return result
        except Exception as e:
            print(f'failed to get_fof_transit_money <err_msg> {e} from {AssetInfo.__tablename__}')

    def get_product_benchmark(self):
        try:
            result = []
            df = self.get_asset_info()
            df = df[df['asset_name'].isin(['大盘股','中证500','创业板指','上证50'])]
            for r in df.itertuples():
                result.append([r.real_id, r.real_name])
            return result
        except Exception as e:
            print(f'failed to get_fof_transit_money <err_msg> {e} from {AssetInfo.__tablename__}')

    

    
    
    

    def get_index_and_asset_price(self, index_list, begin_date, end_date):
        try:
            index_id_list = [i for i in index_list if '-' not in i]
            asset_id_list = [i for i in index_list if '-' in i]
            index_price = pd.DataFrame()
            asset_price = pd.DataFrame()
            if len(index_id_list) > 0:
                #print(f'index_ids {index_id_list}')
                index_price = self.get_index_price_dt(index_list=index_id_list, start_date=begin_date, end_date=end_date)
                index_price = index_price.pivot_table(index='datetime',columns='index_id',values='close')
            if len(asset_id_list) > 0:
                #print(f'asset_id_list {asset_id_list}')
                asset_price = self.get_asset_price_dt_back_with_asset_id(asset_list=asset_id_list,begin_date=begin_date-datetime.timedelta(days=10),end_date=end_date)
            df = pd.concat([index_price,asset_price],join='outer', axis=1).sort_index().ffill().reindex(index_price.index)
            return df
        except Exception as e:
            print(f'failed to asset price <err_msg> {e} from basic.get_index_and_asset_price')
            return pd.DataFrame()
            
    def get_index_and_asset_name_dic(self, index_list):
        try:
            index_id_list = [i for i in index_list if '-' not in i]
            asset_id_list = [i for i in index_list if '-' in i]
            name_dic = {}
            asset_dic = {}
            if len(index_id_list) > 0:
                index_info = self.get_index_info(index_id_list)
                name_dic = index_info.set_index('index_id')['desc_name'].to_dict()
            if len(asset_id_list) > 0:
                asset_info = self.get_asset_info(asset_id_list=asset_id_list)
                asset_dic = asset_info.set_index('asset_id')['real_name'].to_dict()
                name_dic.update(asset_dic)
            return name_dic

        except Exception as e:
            print(f'failed to asset price <err_msg> {e} from basic.get_index_and_asset_info')
            return {}

    def data_resample_monthly_nav(self, df, rule='1M'):
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).last()
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df
            
    def market_size_df(self,begin_date, end_date, time_para, codes=['cninfo_smallcap','cni_largec','cni_midcap']):
        try:
            begin_date, end_date = RawDataApi().get_date_range(time_para, begin_date, end_date)
            index_price = self.get_index_price_dt(start_date=begin_date,end_date=end_date,index_list=codes)
            index_info = self.get_index_info(codes)
            index_price = index_price.pivot_table(index='datetime',columns='index_id',values='close').dropna()
            dic = index_info.set_index('index_id').to_dict()['desc_name']
            index_price = index_price / index_price.iloc[0] - 1
            index_price = index_price.rename(columns=dic)
            index_price.index.name = 'datetime'
            return index_price

        except Exception as e:
            print(f'failed to asset price <err_msg> {e} from basic.market_size_df')

    def get_main_index_future_diff_yearly(self, begin_date, end_date, time_para, codes=['EMM00597141','EMM00597142','EMM00597143']):
        try:
            
            def is_third_friday(d):
                return d.weekday() == 4 and 15 <= d.day <= 21
            # 当月连续合约换仓天数
            def this_months_ex(x):
                return exchange_dts[exchange_dts >= x][0]
            # 下月连续合约换仓天数
            def next_months_ex(x):
                return exchange_dts[exchange_dts >= datetime.date(year=x.year,month=x.month,day=28)][0]
            # 下季度连续合约换仓天数
            def this_season_ex(x):
                return exchange_dts[(exchange_dts >= x) & ([i.month in [3,6,9,12] for i in exchange_dts])][1]

            begin_date, end_date = RawDataApi().get_date_range(time_para, begin_date, end_date)

            # begin_date = max(begin_date, datetime.date(2020,1,1))
            codes = ['EMM00597141','EMM00597142','EMM00597143']
            df = RawDataApi().get_stock_index_future_diff_base(begin_date,end_date,codes)
            index_price = self.get_index_price_dt(start_date=begin_date,index_list=['csi500'])
            index_price = index_price.pivot_table(columns='index_id',values='close',index='datetime')
            df = df.join(index_price)
            for code_i in codes:
                df[code_i] = df[code_i] / df.csi500

            dt1 = [i for i in df.index if is_third_friday(i)]
            dt2 = []
            for i in range(200):
                _d = df.index.values[-1] + datetime.timedelta(i)
                if is_third_friday(_d):
                    dt2.append(_d)
            exchange_dts = np.array(dt1 + dt2)
            exchange_dts = np.array([i - datetime.timedelta(days=4) for i in exchange_dts])    
                
            df.loc[:,'当月连续剩余交易日'] = df.index.map(lambda x: (this_months_ex(x) - x).days + 4)
            df.loc[:,'下月连续剩余交易日'] = df.index.map(lambda x: (next_months_ex(x) - x).days + 4)
            df.loc[:,'下季连续剩余交易日'] = df.index.map(lambda x: (this_season_ex(x) - x).days + 4)

            df['EMM00597141'] = df['EMM00597141'] * 250 / df['当月连续剩余交易日']
            df['EMM00597142'] = df['EMM00597142'] * 250 / df['下月连续剩余交易日']
            df['EMM00597143'] = df['EMM00597143'] * 250 / df['下季连续剩余交易日']

            dic = {
                        'EMM00597141':'IC当月连续年化贴水率',
                        'EMM00597142':'IC下月连续年化贴水率',
                        'EMM00597143':'IC下季连续年化贴水率',
                    }
            df = df[codes].rename(columns=dic)
            df.index.name = 'datetime'
            return df

        except Exception as e:
            print(f'failed to asset price <err_msg> {e} from basic.get_main_index_future_diff_yearly')
    
    def get_industry_list(self):
        try:
            df = RawDataApi().get_em_industry_info(ind_class_type=1)[['em_id','ind_name']]
            em_id_list = df.em_id.to_list()
            df = self.get_index_info_by_em_id(em_id_list)
            info_list = [(r.index_id,r.desc_name.replace('(申万)','')) for r in df.itertuples()]
            return info_list
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from get_em_industry_info')
            return []

    def get_industry_price(self, begin_date, end_date, time_para, industry_list):
        try:
            begin_date, end_date = RawDataApi().get_date_range(time_para, begin_date, end_date)
            df_info = self.get_index_info(industry_list)
            info_dic = {r.index_id : r.desc_name.replace('(申万)','') for r in df_info.itertuples()}
            df = self.get_index_price_dt(start_date=begin_date,end_date=end_date,index_list=industry_list)
            df = df.pivot_table(columns='index_id',values='close',index='datetime').dropna()
            df = df / df.iloc[0]
            df = df.rename(columns=info_dic)
            return df

        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from get_industry_price')
            return pd.DataFrame([])

    def industry_recent_rate(self, industry_list, year:int=None, weekly:bool=False):
        try:
            df_info = self.get_index_info(industry_list)
            desc_name_dic = {r.index_id : r.desc_name.replace('(申万)','') for r in df_info.itertuples()}
            index_price = self.get_index_price_dt(start_date='20091231',index_list=industry_list).pivot_table(index='datetime',columns='index_id',values='close')
            if year is None:
                # 年度
                index_price_y = self.data_resample_monthly_nav(index_price.bfill(),rule='12M')
                index_ret_year = index_price_y.pct_change(1).dropna()
                td = index_ret_year.index
                td = [str(i.year) for i in td]
                index_ret = index_ret_year.copy()
            else:
                index_price_m = index_price.loc[datetime.date(year-1,12,31):datetime.date(year,12,31)]
                index_price_m = self.data_resample_monthly_nav(index_price_m,rule='1M').bfill()
                index_ret_m = index_price_m.pct_change(1).dropna()
                td = index_ret_m.index
                td = [str(i.year) + str(i.month).zfill(2) for i in td ]
                index_ret = index_ret_m.copy()
            if weekly:
                index_price_w = self.data_resample_monthly_nav(index_price,rule='1W').bfill()
                index_ret = index_price_w.pct_change(1).dropna().tail(12)
                td = index_ret.index
                td = [i.strftime('%Y%m%d') for i in td]
            last_day = td[-1]
            index_ret.index = td
            index_ret = index_ret.T.sort_values(last_day, ascending=False).T
            index_ret.loc['均值',:] = index_ret.mean()
            index_ret = index_ret.round(4)*100
            index_ret = index_ret.rename(columns=desc_name_dic)
            return index_ret.T
        except Exception as e:
            print(f'failed to asset price <err_msg> {e} from basic.get_industry_recent_ret')

    def get_north_cap_data(self, begin_date, end_date, time_para, index_id):
        try:
            raw_api = RawDataApi()
            begin_date, end_date = raw_api.get_date_range(time_para, begin_date, end_date)
            index_price = self.get_index_price_dt(start_date=begin_date,end_date=end_date,index_list=[index_id])
            index_info = self.get_index_info(index_list=[index_id])
            desc_name = index_info.set_index('index_id').to_dict()['desc_name']
            desc_name['north_cap'] = '北向资金'
            north_cap = raw_api.get_north_cap(begin_date, end_date)
            index_price = index_price.pivot_table(index='datetime',values='close',columns='index_id')
            north_cap =north_cap.set_index('datetime')[['north_cap']]
            df = index_price.join(north_cap).rename(columns=desc_name)
            return df
        except Exception as e:
            print(f'failed to asset price <err_msg> {e} from basic.get_north_cap_data')

    def get_south_cap_data(self, begin_date, end_date, time_para, index_id):
        try:
            raw_api = RawDataApi()
            begin_date, end_date = raw_api.get_date_range(time_para, begin_date, end_date)
            index_price = self.get_index_price_dt(start_date=begin_date,end_date=end_date,index_list=[index_id])
            index_info = self.get_index_info(index_list=[index_id])
            desc_name = index_info.set_index('index_id').to_dict()['desc_name']
            desc_name['south_cap'] = '南向资金'
            north_cap = raw_api.get_south_cap(begin_date, end_date)
            index_price = index_price.pivot_table(index='datetime',values='close',columns='index_id')
            north_cap =north_cap.set_index('datetime')[['south_cap']]
            df = index_price.join(north_cap).rename(columns=desc_name)
            return df
        except Exception as e:
            print(f'failed to asset price <err_msg> {e} from basic.get_south_cap_data')

    def get_side_cap_index(self):
        index_list = [{'id':'sse50','name':'上证50'},{'id':'hs300','name':'沪深300'},{'id':'csi500','name':'中证500'},{'id':'csi1000','name':'中证1000'},{'id':'hsi','name':'恒生指数'}]
        return index_list

    def get_side_cap_data(self, direction, begin_date, end_date, time_para, index_id):
        try:
            if direction == '北向资金':
                return self.get_north_cap_data(begin_date, end_date, time_para, index_id).dropna()
            else:
                return self.get_south_cap_data(begin_date, end_date, time_para, index_id).dropna()
        except Exception as e:
            print(f'failed to asset price <err_msg> {e} from basic.get_side_cap_data')

    def get_ppi(self, begin_date, end_date, time_para):
        try:
            raw = RawDataApi()
            begin_date, end_date = raw.get_date_range(time_para, begin_date, end_date)
            df = raw.get_em_macroeconomic_monthly(codes=['PPI'],start_date=begin_date,end_date=end_date)
            df = df.drop(columns=['_update_time','codes']).rename(columns={'datetime':'datetime','value':'同比'})
            df.loc[:,'总增长'] = df.同比.cumsum()
            if df.shape[0] > 100:
                idxs = df.index.tolist()
                _idxs = [ i for i in idxs if i%3 == 0]
                diff = idxs[-1] - _idxs[-1]
                idxs = [i + diff for i in _idxs]
                df = df.loc[idxs]
                        
            index_price = self.get_index_price_dt(start_date=begin_date,end_date=end_date,index_list=['hs300'])
            index_price = index_price.set_index('datetime')[['close']].loc[df.datetime.values[0]:df.datetime.values[-1]]
            index_ret = round(index_price.close[-1] / index_price.close[0] - 1,3)
            text = ['反映一定时期内全部工业产品出厂价格总水平的变动趋势和变动幅度的相对数。',
                    '衡量工业企业产品出厂价格变动趋势和变动程度的指数，是反映某一时期生产领域价格变动情况的重要经济指标。',
                    '由生产成本、利润和税金三部分组成，它是工业产品进入流通领域的最初价格。',
                    '指数比预期数值高时，表明有通货膨胀的风险；指数比预期数值低时，则表明有通货紧缩的风险。']
            data = {
                'data':df,
                '沪深300总增长':index_ret,
                '介绍':text,
            }
            return data
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from basic.get_ppi')
            return pd.DataFrame([])
    
    def get_pmi(self, begin_date, end_date, time_para):
        try:
            raw = RawDataApi()
            begin_date, end_date = raw.get_date_range(time_para, begin_date, end_date)
            l = ['CX_PMI','PMI','NonPMI']
            df = raw.get_em_macroeconomic_monthly(codes=l)
            info = raw.get_em_macroeconomic_info(codes=l)
            name_dic = info.set_index('codes').to_dict()['desc_name']
            name_dic['CX_PMI'] = '财新PMI'
            df = df.pivot_table(index='datetime',columns='codes',values='value').dropna().rename(columns=name_dic)
            pct_chg = lambda x: x[-1] / x[0] - 1
            df.loc[:,'财新PMI同比'] = df.财新PMI.rolling(window=13).apply(pct_chg)
            df.loc[:,'非制造业PMI同比'] = df.非制造业PMI.rolling(window=13).apply(pct_chg)
            df.loc[:,'PMI同比'] = df.PMI.rolling(window=13).apply(pct_chg)
            df = df.loc[begin_date:end_date]
            df.index.name = 'datetime'
            text = ['衡量一个国家制造业的 “体检表”，是制造业在生产、新订单、商品价格、存货、 雇员、订单交货、新出口订单和进口等八个方面状况的指数。',
                    'PMI指数50为荣枯分水线。PMI略大于50，说明经济在缓慢前进，PMI略小于50说明经济在慢慢走向衰退。']
            data = {
                'data':df,
                '介绍':text,
            }
            return data
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from basic.get_pmi')
            return pd.DataFrame([])

    def get_gdp(self, begin_date, end_date, time_para):
        try:
            l = ['gdp_cum_yoy',
                'gdp_i1_c_yoy',
                'gdp_i1_s_w',
                'gdp_i2_c_yoy',
                'gdp_i2_s_w',
                'gdp_i3_c_yoy',
                'gdp_i3_s_w',
                ]
            raw = RawDataApi()
            begin_date, end_date = raw.get_date_range(time_para, begin_date, end_date)
            df = raw.get_em_macroeconomic_monthly(codes=l,start_date=begin_date,end_date=end_date)
            info = raw.get_em_macroeconomic_info(codes=l)
            desc_dic = info.set_index('codes').to_dict()['desc_name']
            df = df.pivot_table(index='datetime',columns='codes',values='value')
            df = df.rename(columns=desc_dic)
            name_dic = {
                'GDP：累计同比':'总GDP同比增速',
                'GDP第一产业累计同比':'第一产业同比增速',
                '第一产业GDP当季占比':'第一产业当季贡献',
                'GDP第二产业累计同比':'第二产业同比增速',
                '第二产业GDP当季占比':'第二产业当季贡献',
                'GDP第三产业累计同比':'第三产业同比增速',
                '第三产业GDP当季占比':'第三产业当季贡献',
            }
            df = df.rename(columns=name_dic).ffill()
            text = [
                '国内生产总值，简称GDP，是指在一定时期内（一个季度或一年），全国经济中所生产出的全部最终产品和劳务的价值。 季度更新每年4月、7月、10月、来年1月，15号左右',
                '第一产业主要指生产食材以及其它一些生物材料的产业，包括种植业、林业、畜牧业、水产养殖业等直接以自然物为生产对象的产业。',
                '第二产业主要指加工制造产业，利用自然界和第一产业提供的基本材料进行加工处理。',
                '第三产业是指第一、第二产业以外的其他行业，范围比较广泛，主要包括交通运输业、通讯产业、商业、餐饮业、金融业、教育、公共服务等非物质生产部门。',
            ]
            data = {
                'data':df,
                'text':text
            }
            return data

        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from basic.get_gdp')
            return pd.DataFrame([])

    def get_cpi(self, begin_date, end_date, time_para):
        try:
            l = ['CPI','CPI_mom']
            raw = RawDataApi()
            begin_date, end_date = raw.get_date_range(time_para, begin_date, end_date)
            df = raw.get_em_macroeconomic_monthly(codes=l,start_date=begin_date,end_date=end_date)
            df = df.pivot_table(columns='codes',values='value',index='datetime').dropna()
            _df = (df['CPI_mom']/100 + 1).cumprod()
            df.loc[:,'CPI估算指数'] = _df / _df.iloc[0]
            df = df.drop(columns=['CPI_mom'])
            index_price = BasicDataApi().get_index_price_dt(start_date=df.index[0],end_date=df.index[-1],index_list=['hs300'])
            index_price = index_price.pivot_table(index='datetime',values='close',columns='index_id')
            index_price = index_price.reindex(index_price.index.union(df.index).sort_values()).ffill()
            df = df.join(index_price).dropna().rename(columns={'hs300':'沪深300'}).round(2)
            text = ['消费者物价指数（consumer price index），又名居民消费价格指数，简称CPI。',
                    '反映居民家庭购买的消费品和服务项目的价格水平变动情况，通常作为观察通货膨胀水平的重要指标。',
                    '由以下8大类构成：食品烟酒、衣着、居住、生活用品及服务、交通和通信、教育文化和娱乐、医疗保健、其他用品和服务。',]
            data = {
                'data':df,
                'text':text,
            }
            
            return data
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from basic.get_cpi')
            return pd.DataFrame([])

    def get_ah_prem(self, begin_date, end_date, time_para, frequency):
        try:
            raw = RawDataApi()
            index_id = 'em_ah_prem'
            begin_date, end_date = raw.get_date_range(time_para, begin_date, end_date)
            index_price = BasicDataApi().get_index_price_dt(start_date=begin_date,end_date=end_date,index_list=[index_id])
            index_price = index_price.set_index('datetime')[['open','high','low','close']]
            title = 'AH股溢价指数'
            if frequency != '1D':
                index_price = index_price.set_axis(pd.to_datetime(index_price.index), inplace=False).resample(frequency).last()
                index_price.index = [i.date() for i in index_price.index]
                index_price.index.name = 'datetime'
            
            index_price.loc[:,'MA5'] = index_price.close.rolling(window=5,min_periods=1).mean() 
            index_price.loc[:,'MA10'] = index_price.close.rolling(window=10,min_periods=1).mean() 
            index_price.loc[:,'MA30'] = index_price.close.rolling(window=30,min_periods=1).mean() 
            index_price.loc[:,'MA60'] = index_price.close.rolling(window=60,min_periods=1).mean() 
            data = {
                'data':index_price,
                'title':title
            }
            return data

        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from basic.get_ah_prem')
            return pd.DataFrame([])

    def get_us_bond_rate(self, begin_date, end_date, time_para):
        try:
            rate_dic = {
                'US_TB_10Y':'美国十年期国债收益率',
                'US_TB_2Y':'美国两年期国债收益率',
                'US_TB_3M':'美国三月期国债收益率',
                'US_TB_5Y':'美国五年期国债收益率',
                'TB_10Y':'中国十年国债收益率',
                'TB_1Y':'中国一年国债收益率',
                'TB_3Y':'中国三年国债收益率',
                'TB_5Y':'中国五年国债收益率',
            }
            raw = RawDataApi()
            begin_date, end_date = raw.get_date_range(time_para, begin_date, end_date)
            us_debt = raw.get_em_macroeconomic_daily(codes=rate_dic.keys(),start_date=begin_date, end_date=end_date)
            df = us_debt.pivot_table(index='datetime',values='value',columns='codes').dropna()
            df = df.rename(columns=rate_dic)
            df.index.name = 'datetime'
            data = {
                'data':df,
                'text':['选取美国和中国国债名义利率']
            }
            return data
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from basic.get_us_bond_rate')
            return pd.DataFrame([])

    def get_usd_index(self, begin_date, end_date, time_para):
        try:
            raw = RawDataApi()
            begin_date, end_date = raw.get_date_range(time_para, begin_date, end_date)
            rates = RawDataApi().get_raw_cm_index_price_df(start_date=begin_date,end_date=end_date)
            rates = rates[['datetime','usd_central_parity_rate']].rename(columns={'usd_central_parity_rate':'美元人民币汇率'})
            data = {
                'data':rates,
                'text':['选取中美汇率中间价'],
            }
            return data
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from basic.get_usd_index')
            return pd.DataFrame([])
    
    def get_rzrqye(self, begin_date, end_date, time_para):
        try:
            raw = RawDataApi()
            begin_date, end_date = raw.get_date_range(time_para, begin_date, end_date)
            code_list = ['rq_ye','rz_ye']
            asset_info = raw.get_em_macroeconomic_info(codes=code_list)
            df = raw.get_em_macroeconomic_daily(codes=code_list,start_date=begin_date,end_date=end_date)
            df = df.pivot_table(index='datetime',columns='codes',values='value')
            df = df.rename(columns=asset_info.set_index('em_codes').to_dict()['desc_name'])
            df = df/1e8
            index_list = ['hs300','csi500','csi1000']
            index_info = BasicDataApi().get_index_info(index_list=index_list)
            index_price = BasicDataApi().get_index_price_dt(start_date=begin_date, end_date=end_date, index_list=index_list)
            index_price = index_price.pivot_table(index='datetime',columns='index_id',values='total_turnover').ffill()
            index_price = index_price.dropna()
            index_name = index_info.set_index('index_id').to_dict()['desc_name']
            index_price = index_price.rename(columns=index_name)/1e8
            index_price = index_price.set_axis(pd.to_datetime(index_price.index), inplace=False).resample('W-FRI').sum()
            index_price.index = [i.date() for i in index_price.index]
            index_price.index.name = 'datetime'
            index_price = index_price.join(df).dropna()
            data = {
                'data':index_price,
                'text':['股指交易量是周度加和,单位亿元'],
            }
            return data

        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from basic.get_rzrqye')
            return pd.DataFrame([])

    def _get_asset_benchmark_info_list(self):
        try:
            index_info = self.get_index_info()
            l = ['恒生指数有限公司', '巨潮资讯', '中证指数有限公司','深圳证券信息有限公司', '深圳证券交易所', '上海证券交易所',
                '东方财富','标准普尔公司','上海申银万国证券研究所有限公司']
            index_info = index_info[index_info.maker_name.isin(l)]
            result = [[r.index_id, r.desc_name] for r in index_info[['index_id','desc_name']].itertuples()]
            return result

        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from basic.get_asset_benchmark_info_list')
            return []

    def get_asset_benchmark_info_list(self):
        try:
            asset_type_list = ['大类资产', '期货指数', '其他地区', '行业指数', '市场指数', '策略指数', '股票指数',
                '风格指数', '主题指数']
            asset_info = BasicDataApi().get_asset_info_by_type(asset_type=asset_type_list)
            index_info = BasicDataApi().get_index_info(index_list=asset_info.real_id.tolist())
            existed_id_info = asset_info[asset_info.real_id.isin(index_info.index_id)].copy()
            not_exsited_if_info = asset_info[~asset_info.real_id.isin(index_info.index_id)]
            index_info.loc[:,'others'] = index_info.desc_name +'(' + index_info.em_id + ')'
            index_others_dic = index_info.set_index('index_id')['others'].dropna().to_dict()
            index_desc_dic = index_info.set_index('index_id')['desc_name'].to_dict()
            existed_id_info.loc[:,'others'] = existed_id_info.real_id.map(lambda x: index_others_dic[x] if x in index_others_dic else index_desc_dic[x])
            result1 = [[r.real_id, r.others] for r in existed_id_info.itertuples()]
            result2 = [[r.asset_id, r.real_name] for r in not_exsited_if_info.itertuples()]
            result = result1+result2
            return result

        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from basic.get_asset_benchmark_info_list')
            return []
        
    def get_asset_benchmark_info_with_layers(self):
        try:
            asset_type_list = ['大类资产', '期货指数', '其他地区', '行业指数', '市场指数', '策略指数', '股票指数',
                '风格指数', '主题指数']
            asset_info = BasicDataApi().get_asset_info_by_type(asset_type=asset_type_list)
            index_info = BasicDataApi().get_index_info(index_list=asset_info.real_id.tolist())
            existed_id_info = asset_info[asset_info.real_id.isin(index_info.index_id)].copy()
            not_exsited_if_info = asset_info[~asset_info.real_id.isin(index_info.index_id)]
            index_info.loc[:,'others'] = index_info.desc_name +'(' + index_info.em_id + ')'
            index_others_dic = index_info.set_index('index_id')['others'].dropna().to_dict()
            index_desc_dic = index_info.set_index('index_id')['desc_name'].to_dict()
            existed_id_info.loc[:,'others'] = existed_id_info.real_id.map(lambda x: index_others_dic[x] if x in index_others_dic else index_desc_dic[x])
            existed_dic = {
                'real_id':'id',
                'others':'name',
                'asset_type':'asset_type',
            }
            existed_id_info = existed_id_info[existed_dic.keys()].rename(columns=existed_dic)
            not_existed_dic = {
                'asset_id':'id',
                'real_name':'name',
                'asset_type':'asset_type'
            }
            not_exsited_if_info = not_exsited_if_info[not_existed_dic.keys()].rename(columns=not_existed_dic)
            asset_info = pd.concat([existed_id_info,not_exsited_if_info])
            asset_types = asset_info.asset_type.unique().tolist()
            _result = []
            for i in asset_types:
                dic = {
                    'title':i
                }
                _df_i = asset_info[asset_info.asset_type == i]
                _res_details = []
                for r in _df_i.itertuples():
                    _dic = {'asset_id':r.id, 'desc_name':r.name}
                    _res_details.append(_dic)
                dic['items'] = _res_details
                _result.append(dic)
            return _result
            
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from basic.get_asset_benchmark_info_with_layers')
            return []

    def get_asset_name(self, id_list:list=[]):
        try:
            hf_fund_info = self.get_fof_info_strategy_exsit(fof_id_list=id_list)
            fund_info = self.get_fund_info(fund_list=id_list)
            index_info = self.get_index_info(index_list=id_list)
            name_dic = {}
            if not hf_fund_info.empty:
                d1 = hf_fund_info.set_index('fof_id')['fof_name'].to_dict()
                name_dic.update(d1)
            if not fund_info.empty:
                d2 = fund_info.set_index('fund_id')['desc_name'].to_dict()
                name_dic.update(d2)
            if not index_info.empty:
                d3 = index_info.set_index('index_id')['desc_name'].to_dict()
                name_dic.update(d3)
            return name_dic

        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from basic.get_asset_name')
            return {}

    def delete_hedge_fund_email_raw(self, manager_id: str, fof_id: str, fund_id: str, date_list: List[datetime.date]) -> bool:
        with BasicDatabaseConnector().managed_session() as session:
            try:
                session.query(
                    HedgeFundEmailRaw
                ).filter(
                    HedgeFundEmailRaw.manager_id == manager_id,
                    HedgeFundEmailRaw.fof_id == fof_id,
                    HedgeFundEmailRaw.fund_id == fund_id,
                    HedgeFundEmailRaw.datetime.in_(date_list),
                ).delete(synchronize_session=False)
                session.commit()
                return True
            except Exception as e:
                print(f'failed to delete data <err_msg> {e} from {HedgeFundEmailRaw.__tablename__}')
                return False



from sqlalchemy import CHAR, Column, Integer, TEXT, BOOLEAN, text, Index, Enum
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.mysql import DOUBLE, DATE, TINYINT, DATETIME, MEDIUMTEXT, YEAR, SMALLINT, BIGINT, DECIMAL, VARCHAR

from ...constant import SectorType, IndexPriceSource, CodeFeeMode, IndClassType


class Base:
    _update_time = Column('_update_time', DATETIME, nullable=False, server_default=text('CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP'))  # 更新时间


class FOFBase:
    create_time = Column(DATETIME, nullable=False, server_default=text('CURRENT_TIMESTAMP'))  # 创建时间
    update_time = Column(DATETIME, nullable=False, server_default=text('CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP'))  # 更新时间
    is_deleted = Column(BOOLEAN, nullable=False, server_default=text('FALSE'))  # 是否删除 默认不删除


# make this column at the end of every derived table
Base._update_time._creation_order = 9999
Base = declarative_base(cls=Base)

FOFBase.create_time._creation_order = 9997
FOFBase.update_time._creation_order = 9998
FOFBase.is_deleted._creation_order = 9999
FOFBase = declarative_base(cls=FOFBase)

# test库主要是测试库，使用没问题后，转正

class PfIndexInfo(Base):
    '''私募指数信息'''

    __tablename__ = 'pf_index_info'

    index_id = Column(CHAR(20), primary_key=True) # 指数id
    desc_name = Column(CHAR(40)) # 指数中文
    start_date = Column(DATE) # 指数开始日期
    desc = Column(CHAR(255)) # 指数说明

class PfIndexComponent(Base):
    '''私募指数权重表'''

    __tablename__ = 'pf_index_component'

    index_id = Column(CHAR(20), primary_key=True) # 指数id
    datetime = Column(DATE, primary_key=True) # 开始日期
    desc_name = Column(CHAR(128), primary_key=True) # 私募基金名
    wgt = Column(DOUBLE(asdecimal=False)) # 权重

class PfIndexPrice(Base):
    '''私募指数净值表'''

    __tablename__ = 'pf_index_price'

    index_id = Column(CHAR(20), primary_key=True) # 指数id
    datetime = Column(DATE, primary_key=True) # 开始日期
    close = Column(DOUBLE(asdecimal=False)) # 收盘价
    ret = Column(DOUBLE(asdecimal=False)) # 收益率


from http.server import HTTPServer
from http.server import BaseHTTPRequestHandler
from urllib.parse import urlparse, quote, unquote
from xml.etree import ElementTree
import random
import threading
# import ssl
from ...util.config import Configurator
from ...util.struct import WxMsgItem
from ...util import Singleton
from ...util.wx_crypto.WXBizMsgCrypt3 import WXBizMsgCrypt

class MessageCrypt(metaclass=Singleton):

    def __init__(self):
        token = Configurator().config['message_receiver']['wxwork']['token']
        aes_key = Configurator().config['message_receiver']['wxwork']['aes_key']
        receive_id = Configurator().config['message_receiver']['wxwork']['corp_id']
        self._msg_crypt = WXBizMsgCrypt(token, aes_key, receive_id)
    
    def get_msg_crypt(self):
        return self._msg_crypt

class WxHTTPRequestHandler(BaseHTTPRequestHandler):

    def do_GET(self):
        query = urlparse(self.path).query
        query_components = dict(qc.split("=") for qc in query.split("&"))
        msg_signature = unquote(query_components["msg_signature"])
        timestamp = unquote(query_components["timestamp"])
        nonce = unquote(query_components["nonce"])
        echostr = unquote(query_components["echostr"])

        ret, rsp_echo = MessageCrypt().get_msg_crypt().VerifyURL(msg_signature, timestamp, nonce, echostr)
        self.send_response(200)
        self.send_header('Content-Type', 'text/plain; charset=utf-8')
        # self.send_header('Access-Control-Allow-Origin', '*')
        self.end_headers()
        self.wfile.write(rsp_echo)

    def do_POST(self):
        query = urlparse(self.path).query
        query_components = dict(qc.split("=") for qc in query.split("&"))
        msg_signature = unquote(query_components["msg_signature"])
        timestamp = unquote(query_components["timestamp"])
        nonce = unquote(query_components["nonce"])

        # Get the size of post data
        content_length = int(self.headers['Content-Length'])
        # Get the post data
        post_data = self.rfile.read(content_length)

        ret, xml_content = MessageCrypt().get_msg_crypt().DecryptMsg(post_data, msg_signature, timestamp, nonce)
        xml_root = ElementTree.fromstring(xml_content)
        from_user = xml_root.find('FromUserName').text
        msg_type = xml_root.find('MsgType').text
        content = xml_root.find('Content').text
        msg_id = xml_root.find('MsgId').text
        agent_id = xml_root.find('AgentID').text

        if msg_type == 'text':
            data = {
                'msgid': msg_id,
                'from': f'wo{agent_id}',
                'tolist': [from_user],
                'msgtype': msg_type,
                'msgtime': int(timestamp) * 1000, # convert to ms
                'text': {'content': content}
            }
            msg = WxMsgItem(data)
            print(msg)

            with self.server.msg_mutex:
                self.server.msgs.append(msg)

        self.send_response(200)
        self.send_header('Content-Type', 'text/plain; charset=utf-8')
        self.end_headers()

class MyHTTPServer(HTTPServer):
    def __init__(self, server_address, RequestHandlerClass):
        super().__init__(server_address, RequestHandlerClass)
        self.msg_mutex = threading.Lock()
        self.msgs = []

    def get_msg(self, count):
        with self.msg_mutex:
            result = self.msgs[:count]
            self.msgs = self.msgs[count:]
        return result

class MessageReceiver(object):
    def __init__(self):
        # keyfile = Configurator().config['cert']['keyfile']
        # certfile = Configurator().config['cert']['certfile']
        port = Configurator().config['message_receiver']['wxwork']['port']

        self.server = MyHTTPServer(('0.0.0.0', port), WxHTTPRequestHandler)
        # server.socket = ssl.wrap_socket(server.socket, 
        #     keyfile=keyfile, 
        #     certfile=certfile, server_side=True)

    def run(self):
        self.server.serve_forever()

    def stop(self):
        self.server.shutdown()

    def get_msg(self, count):
        return self.server.get_msg(count)

if __name__ == '__main__':
    mr = MessageReceiver()
    print('Starting server, use <Ctrl-C> to stop')
    mr.run()



from string import digits
from ...api.research_api import *
from ...api.raw import *
from .derived_data_helper import DerivedDataHelper

class FutureTypeProcessor:

    future_type_dic = {
        '股指期货':['中证500股指期货','沪深300股指期货','上证50股指期货'],
        '国债期货':['5年期国债期货','10年期国债期货','2年期国债期货',],
        '贵金属':['沪银','沪金'],
        '有色金属':['沪铜','沪铝','沪锌','沪铅','镍','锡'],
        '黑色金属':['螺纹钢','铁矿石','热卷','不锈钢','线材','硅铁','锰硅'],
        '能源化工':['橡胶','20号胶','塑料','PTA','短纤','PVC','乙二醇','苯乙烯','尿素','纯碱','胶合板','沥青','纤维板','玻璃','燃油','甲醇','液化石油气','聚丙烯','中质含硫原油','纸浆','焦炭','焦煤','线材','低硫燃料油','郑煤'],
        '农产品':['豆粕','玉米','豆一','玉米淀粉','强麦','普麦','粳米','早籼','晚籼','苹果','豆二','郑棉','红枣','棉纱','生猪','菜油','棕榈','花生','菜粕','菜籽','白糖','鸡蛋','粳稻','豆油',],        
    }

    def __init__(self, data_helper = DerivedDataHelper()):
        self._data_helper = data_helper
        self.raw_api =  RawDataApi()
        self.derived_api = DerivedDataApi()
        self.research_api = ResearchApi()

    def init(self):
        self.future_info = self.raw_api.get_em_future_info()
        self.future_type = {vi : k  for k, v in self.future_type_dic.items() for vi in v}

    def remove_digit(self, s):
        remove_digits = str.maketrans('', '', digits)
        return s.translate(remove_digits)

    def process(self):
        result = []
        for product_i, type_i in self.future_type.items():
            trading_code = self.future_info[self.future_info.trans_type == product_i].trading_code.values[-5]
            trading_code = self.remove_digit(trading_code)
            dic = {
                'future_type_id': trading_code,
                'desc_name': product_i,
                'product_type':type_i,
            }
            result.append(dic)
        df = pd.DataFrame(result)   
        df.to_sql('future_type_classification', DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')
    
class FutureMainForcePrice:

    def __init__(self, data_helper = DerivedDataHelper()):
        self._data_helper = data_helper
        self.raw_api = RawDataApi()
        self.derived_api =  DerivedDataApi()
        self.research_api = ResearchApi()   

    def init(self):
        self.future_info = self.raw_api.get_em_future_info()
        self.future_type = self.derived_api.get_future_type_classification()
        self.future_price = self.research_api.get_future_quote()
        self.trade_dates = np.array(self.research_api.get_tdates())
        self.trade_dates_wti = np.array(self.research_api.get_tdates(exch='nyse'))

    def get_main_ticker(self, x):
        # 获取东财主力连续代码
        return x + '0'
        
    def fill_main_future(self, col_name, df):
        l = df[df[col_name] < df[col_name].shift(1)].index.tolist()
        df.loc[l] = None
        df = df.ffill()   
        return df

    def get_main_force(self, future_name, future_type_id, amount_over_days, switch_days):
        # 用收益率做主力连续合约
        ## 去掉东财主连代码
        future_info_i = self.future_info[self.future_info.trans_type == future_name]
        s = future_info_i.trading_code.values[0]
        main_type_name = self.get_main_ticker(future_name)
        future_info_i = future_info_i[future_info_i.trading_code != main_type_name]

        ## 去掉东财股指期货主连
        future_info_i = future_info_i[~future_info_i.trading_code.isin(['ICM','IC00C1','IC00C2','IC00C3','IC00C4','IFM','IF00C1','IF00C2','IF00C3','IF00C4','IHM','IH00C1','IH00C2','IH00C3','IH00C4','CL00Y'])]
        em_id_list = future_info_i.em_id.tolist()
        size = future_info_i[future_info_i.trans_type == future_name].contract_mul.values[0]

        ## 主连合约不允许更改远期合约到近期合约
        future_price_i = self.future_price[self.future_price.em_id.isin(em_id_list)].copy()
        future_price_i['vwap'] = future_price_i['uni_amount'] / future_price_i['uni_volume'] / size
        future_volume = future_price_i.pivot_table(index='datetime', columns='em_id', values='uni_volume')
        main_ticker_df = future_volume.idxmax(axis=1).to_frame()
        col_name = 'main_ticker'
        main_ticker_df.columns = [col_name]
        main_ticker_df = self.fill_main_future(col_name, main_ticker_df)
        main_ticker_df = self.fill_main_future(col_name, main_ticker_df)

        ## 设置交易量连续超过几天后更换合约
        main_ticker_df.loc[:,'last_days_main_ticker'] = main_ticker_df.main_ticker.shift(amount_over_days-1)
        res = []
        for r in main_ticker_df.itertuples():
            dic = {'datetime':r.Index}
            if r.last_days_main_ticker is np.nan or r.last_days_main_ticker == r.main_ticker:
                dic['ticker'] = r.main_ticker
            elif r.last_days_main_ticker != r.main_ticker:
                dic['ticker'] = r.last_days_main_ticker
            res.append(dic)
        ticker_info = pd.DataFrame(res)

        ## 次主力合约不允许更改远期合约到近期合约
        _future_volume = future_volume.copy()
        for r in ticker_info.itertuples():
            _future_volume.loc[r.datetime,r.ticker] = 0
        sub_main_ticker_df = _future_volume.idxmax(axis=1).to_frame()
        col_name = 'sub_ticker'
        sub_main_ticker_df.columns = [col_name]
        sub_main_ticker_df = self.fill_main_future(col_name, sub_main_ticker_df)
        sub_main_ticker_df = self.fill_main_future(col_name, sub_main_ticker_df)
        sub_main_ticker_df = self.fill_main_future(col_name, sub_main_ticker_df)

        ## 设置更换主连后多少天之后更换新收益率 含下一日主力合约
        future_close = future_price_i.pivot_table(index='datetime', columns='em_id', values='close')
        ticker_info = ticker_info.set_index('datetime').join(sub_main_ticker_df).reset_index()
        if future_type_id != 'CL':
            next_dt = self.trade_dates[self.trade_dates>ticker_info.datetime.values[-1]][0]
            _trade_dates = self.trade_dates[(self.trade_dates >= ticker_info.datetime.values[0]) & (self.trade_dates <= ticker_info.datetime.values[-1])]
        
        else:
            next_dt = self.trade_dates_wti[self.trade_dates_wti>ticker_info.datetime.values[-1]][0]
            _trade_dates = self.trade_dates_wti[(self.trade_dates_wti >= ticker_info.datetime.values[0]) & (self.trade_dates_wti <= ticker_info.datetime.values[-1])]
        ticker_info = ticker_info.append(pd.DataFrame([{'datetime':next_dt}])).reset_index(drop=True)
        ticker_info.ticker = ticker_info.ticker.shift(switch_days)
        ticker_info = ticker_info.set_index('datetime').reindex(_trade_dates).ffill().reset_index()
        ticker_info = ticker_info.dropna()
        ## 填充收益率
        future_ret = future_close.pct_change(periods=1,fill_method=None)
        for r in ticker_info.itertuples():
            if r.datetime in future_ret.index:
                ticker_info.loc[r.Index,'ret'] = future_ret.loc[r.datetime,r.ticker]
                ticker_info.loc[r.Index,'sub_ret'] = future_ret.loc[r.datetime,r.sub_ticker]
        ticker_info['future_type_id'] = future_type_id
        return ticker_info

    def process(self):
        try:
            self.init()
            result = []
            amount_over_days = 1
            switch_days = 1
            for r in self.future_type.itertuples():
                try:
                    future_name = r.desc_name
                    future_type_id = r.future_type_id
                    result_i = self.get_main_force(future_name, future_type_id, amount_over_days, switch_days)
                    result.append(result_i)
                except:
                    print(future_name, future_type_id, 'boom')
            result_df = pd.concat(result)
           
            # 删除库主力收益和次主力收益为空的
            self.derived_api.delete_null_future_ret()
            # 增量更新
            exsited_df = self.derived_api.get_future_main_chain_ret()
            result_df_doubleidx = result_df.set_index(['datetime','future_type_id'])
            exsited_df_doubleidx = exsited_df.set_index(['datetime','future_type_id'])
            diff = result_df_doubleidx.index.difference(exsited_df_doubleidx.index)
            df_result = result_df_doubleidx.loc[diff].reset_index()
            tds = pd.to_datetime(df_result.datetime)
            tds = [i.date() for i in tds]
            df_result.datetime = tds
            if not df_result.empty:
                df_result.to_sql('future_main_chain_ret', DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')
            return True
        except:
            return False




from ..structure import Order, Direction, OrderType, AssetType, ExchangeID, TransferRequest, Trade, Sell

class StrategyBase:

    def __init__(self, name):
        self._name = name
        pass

    def set_context(self, context):
        self.context = context

    def init(self, params=None):
        pass
 
    def on_tick(self, tick):
        pass

    def on_trade(self, trade):
        pass

    def on_order(self, order_id, order_status):
        pass

    def on_transfer(self, transfer):
        pass
    
    def insert_order(self, ticker, direction, volume, price=None, exchange=ExchangeID.NOT_AVAILABLE, asset_type=None):
        # price is None if this is a market order
        # return order_id if success
        # return None if there is a error
        order = Order(ticker=ticker, direction=direction, volume=volume, exchange=exchange, 
                      asset_type=self.context._asset_types[ticker] if asset_type is None else asset_type,
                      order_type=OrderType.LIMIT if price else OrderType.MARKET, price=price, source=self._name)
        return self.context.insert_order(order)

    def build_order(self, ticker, direction, volume, price=None, exchange=ExchangeID.NOT_AVAILABLE, asset_type=None):        
        order = Order(ticker=ticker, direction=direction, volume=volume, exchange=exchange, 
                      asset_type=self.context._asset_types[ticker] if asset_type is None else asset_type,
                      order_type=OrderType.LIMIT if price else OrderType.MARKET, price=price, source=self._name)
        
        return order

    def cancel_order(self, order_id):
        # input:  order_id (return value of insert_order)
        # output: 
        #         True if the order is cancelled
        #         False if the order is not exists or has already been filled
        return self.context.cancel_order(order_id)

    
    def insert_transfer(self, from_ticker, to_ticker, from_exchange, to_exchange, volume):
        # input: position ticker ('rmb', 'btc', 'usdt' etc.)
        # output:
        #         transfer_id if transfer permitted
        #         None if transfer failed.
        transfer = TransferRequest(from_ticker=from_ticker, to_ticker=to_ticker, from_exchange=from_exchange, to_exchange=to_exchange, volume=volume, source=self._name)
        return self.context.insert_transfer(transfer)

    def get_position(self):
        return self.context.get_position(self._name)

    def get_trade_buy(self, date):
        return [Sell(ti.ticker, ti.trade_volume) for ti in self.context.get_trade(self._name) if (ti.time == date and ti.direction == Direction.BUY)]
            
    def get_market_value(self):
        return self.context.get_market_value(self._name)

    def switch_day(self):
        pass

    def get_sell_day(self, today, period):
        try:
            idx = self.context._data_manager.wrapper.date_list.index(today) - period
            if idx < 0:
                return None
            else:
                return self.context._data_manager.wrapper.date_list[idx]
        except:
            return None


import datetime

from sqlalchemy import CHAR, Column, Integer, Index, BOOLEAN, text, TEXT, Enum, VARCHAR
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.mysql import DOUBLE, DATE, TINYINT, SMALLINT, DATETIME, MEDIUMTEXT, BIGINT, SMALLINT, DECIMAL
from sqlalchemy.sql.selectable import ScalarSelect

from ...constant import SectorType, IndexPriceSource, HoldingAssetType, FundStatus


class Base:
    _update_time = Column('_update_time', DATETIME, nullable=False, server_default=text('CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP'))  # 更新时间

# make this column at the end of every derived table
Base._update_time._creation_order = 9999
Base = declarative_base(cls=Base)



class PfundInstScaleAmac(Base):
    '''私募基金管理人规模'''

    __tablename__ = 'pfund_inst_scale_amac'
    
    party_id = Column(BIGINT, primary_key=True) # 机构内部编码
    
    party_full_name = Column(CHAR(300)) # 机构名称 
    main_fund_type = Column(CHAR(200)) # 管理基金主要类别 
    fund_types = Column(CHAR(200)) # 管理基金主要类别 # todo
    scale = Column(CHAR(50)) # 管理规模 
    inst_id = Column(CHAR(60)) # 源数据id 
    num = Column(Integer) # 变更次数 
    is_new = Column(Integer) # 是否最新
    fetch_time = Column(DATE) # 变动时间 
    update_time = Column(DATETIME) # 更新时间 
    

class PfundManagerInfo(Base):
    '''私募基金经理基本信息表'''

    __tablename__ = 'pfund_manager_info'
    
    person_id = Column(CHAR(60), primary_key=True) # 人员ID
    
    name = Column(CHAR(60)) # 姓名 
    gender = Column(CHAR(60)) # 性别 
    brithday = Column(CHAR(60)) # 生日 
    nationality_cd = Column(CHAR(60)) # 国籍 
    certificate = Column(CHAR(300)) # 证书 
    photo = Column(CHAR(200)) # 照片
    education = Column(Integer) # 最高学历 
    grad_univ = Column(CHAR(60)) # 毕业院校 
    major = Column(CHAR(60)) # 所学专业
    background_desc = Column(TEXT) # 简介
    background = Column(CHAR(60)) # 背景
    mf_to_pf = Column(TINYINT) # 是否公转私/私转公
    year = Column(DECIMAL(4,2)) # 投资年限
    awards_name = Column(CHAR(2000)) # 所获荣誉
    department = Column(CHAR(60)) # 所在部门
    position = Column(CHAR(500)) # 职务
    is_qual = Column(Integer) # 是否具备基金从业资格
    num_dura = Column(BIGINT) # 当前管理的基金数量
    dura_sec_id = Column(MEDIUMTEXT) # 当前管理的基金ID
    dura_sec_name = Column(MEDIUMTEXT) # 当前管理的基金
    party_id = Column(BIGINT) # 任职机构ID
    party_full_name = Column(CHAR(300)) # 任职机构
    rep_sec_id = Column(BIGINT) # 代表产品ID
    rep_sec_name = Column(CHAR(300)) # 代表产品
    return_man_a = Column(DECIMAL(18,10)) # 任职收益
    remark = Column(CHAR(300)) # 备注
    update_time = Column(DATETIME) # 更新时间

class PfundNav(Base):
    '''私募净值'''

    __tablename__ = 'pfund_nav'
    
    security_id = Column(BIGINT, primary_key=True) # 证券内部编码
    publish_date = Column(DATE, primary_key=True) # 数据发布日期 
    end_date = Column(DATE) # 净值日期 
    nav = Column(DECIMAL(19,4)) # 单位净值 
    accum_nav = Column(DECIMAL(19,4)) # 累计净值 
    update_time = Column(DATETIME) # 更新时间 
    adj_nav = Column(DECIMAL(19,4)) # 调整净值 
    nav_unit = Column(DECIMAL(19,4)) # 净值单位 
    return_rate = Column(DECIMAL(19,10)) # 净值收益 
    adj_return_rate = Column(DECIMAL(19,4)) # 调整净值收益 
    
class PfundInfo(Base):
    '''私募基本信息'''
        
    __tablename__ = 'pfund_info'
    
    security_id = Column(BIGINT, primary_key=True) # 证券内部编码 
    establish_date = Column(DATE) # 成立日期 
    pf_style = Column(CHAR(10)) # 产品类型 
    status = Column(CHAR(200)) # 运行状态 
    invest_strategy = Column(CHAR(10)) # 投资策略 
    invest_strategy_child = Column(CHAR(200)) # 子策略 
    duration = Column(Integer) # 存续期限 
    open_date_desc = Column(CHAR(500)) # 开放日 
    invest_consultant = Column(CHAR(50)) # 投资顾问 
    custodian = Column(CHAR(50)) # 托管人 
    issue_platform = Column(CHAR(100)) # 受托人 
    trading_broker = Column(CHAR(50)) # 交易经纪 
    subscription_start_point = Column(Integer) # 认购起点 
    scale_initial = Column(BIGINT) # 初始规模 
    issue_fee = Column(DECIMAL(19,10)) # 认购费率 
    redeem_fee = Column(DECIMAL(19,10)) # 赎回费率 
    management_fee = Column(DECIMAL(19,10)) # 管理费率 
    performanece_return = Column(DECIMAL(19,10)) # 业绩报酬 
    record_cd = Column(CHAR(50)) # 备案编码 
    update_time = Column(DATETIME) # 更新时间 
    record_date = Column(DATE) # 备案日期 
    end_date = Column(DATE) # 到期日 
    nav_freq = Column(CHAR(20)) # 净值发布频率 
    closing_dura_desc = Column(CHAR(200)) # 封闭期 
    issue_loc = Column(CHAR(200)) # 发行地 
    record_status = Column(CHAR(20)) # 备案状态 
    manage_type = Column(TINYINT(3)) # 管理类型 
    manager = Column(CHAR(500)) # 基金经理 
    min_add = Column(CHAR(200)) # 最低追加金额 
    stop_loss_line = Column(DECIMAL(19,4)) # 止损线 
    warn_line = Column(DECIMAL(19,4)) #  预警线
    futures_broker = Column(CHAR(50)) #  期货经纪
    amac_type = Column(CHAR(200)) #  协会公布类型
    apply_fee = Column(DECIMAL(19,4)) # 申购费
    custody_fee = Column(DECIMAL(19,4)) #  托管费
    is_quant = Column(TINYINT(3)) # 是否量化
    is_hedging = Column(TINYINT(3)) #  是否对冲

    
class PfundInstInfo(Base):
    '''投顾机构信息'''
        
    __tablename__ = 'pfund_inst_info'                        

    party_id = Column(BIGINT(20), primary_key=True) # 机构内部编码
    profile = Column(TEXT) # 公司简介
    idea_strategy= Column(TEXT) # 投资理念
    key_person= Column(TEXT) # 核心人物
    update_time= Column(DATETIME) # 更新时间
    record_date= Column(DATE) # 登记日期
    reg_cd= Column(CHAR(20)) # 登记编码
    legal_rep= Column(CHAR(100)) # 法定代表人/执行事务合伙人(委派代表)
    is_qualified= Column(TINYINT(3)) # 法人是否有从业资格
    qualify_way= Column(CHAR(20)) # 法人从业资格认定方式
    emp_num= Column(Integer) # 员工数量
    main_fund_type= Column(CHAR(100)) # 管理基金主要类别
    private_ind_scale= Column(CHAR(20)) # 私募证券基金自主发行规模
    private_con_scale= Column(CHAR(20)) # 私募证券基金顾问管理规模
    pe_scale= Column(CHAR(20)) # 私募股权基金规模
    vc_scale= Column(CHAR(20)) # 创业投资基金规模
    other_scale= Column(CHAR(20)) # 其他私募基金规模
    is_fund_mana= Column(BIGINT) # 是否公募基金管理人
    paa_scale= Column(CHAR(20)) # 私募资产配置基金规模
    record_status= Column(CHAR(20)) # 备案状态
    management_scale= Column(CHAR(100)) # 管理规模区间

class MdInstitution(Base):
    '''公司基本信息'''
        
    __tablename__ = 'md_institution'
    
    party_id = Column(Integer, primary_key=True) # 机构内部ID
    party_full_name = Column(CHAR(100)) # 机构全称 
    party_short_name = Column(CHAR(100)) # 机构简称 
    party_full_name_en = Column(CHAR(200)) # 机构英文全称 
    party_short_name_en = Column(CHAR(200)) # 机构英文简称 
    reg_date = Column(DATE) # 注册日期 
    reg_country_cd = Column(CHAR(3)) # 注册国家代码 
    reg_province = Column(CHAR(100)) # 注册省份 
    reg_city = Column(CHAR(100)) # 注册城市 
    reg_addr = Column(CHAR(200)) # 注册地址 
    reg_cap = Column(DECIMAL(18,4)) # 注册资金 
    reg_cap_curr_cd = Column(CHAR(3)) # 注册资金货币代码 
    office_addr = Column(CHAR(200)) # 办公地址 
    email = Column(CHAR(100)) # 联系邮箱 
    website = Column(CHAR(100)) # 机构网站 
    tel = Column(CHAR(50)) # 联系电话 
    fax = Column(CHAR(50)) # 联系传真 
    legal_entity_id = Column(CHAR(20)) # 工商登记号 
    prime_operating = Column(MEDIUMTEXT) # 主营业务 
    party_nature_cd = Column(CHAR(100)) # 公司性质 
    is_iss_bond = Column(CHAR(100)) # 是否发行债券 
    close_date = Column(DATE) # 机构终止日期 
    inst_status = Column(SMALLINT) # 是否运作 
    dy_use_flg = Column(BIGINT) # DY使用标识 
    update_time = Column(DATETIME) # 更新时间 
    profile = Column(TEXT) # 机构介绍 
    board_secry = Column(CHAR(50)) # 董事会秘书 
    legal_rep = Column(CHAR(50)) # 法定代表人 
    opa_scope = Column(MEDIUMTEXT) # 经营范围 
    gen_mana = Column(CHAR(50)) # 总经理 
    is_iss_mf = Column(Integer) # 是否发行公募基金 

class MdSecurity(Base):
    '''主编码及基本信息'''
        
    __tablename__ = 'md_security'
    
    security_id = Column(Integer, primary_key=True) # 证券内部ID
    ticker_symbol = Column(CHAR(50)) # 交易代码 
    exchange_cd = Column(CHAR(4)) # 交易市场代码 
    sec_full_name = Column(CHAR(100)) # 证券全称 
    sec_short_name = Column(CHAR(100)) # 证券简称 
    cn_spell = Column(CHAR(200)) # 简称拼音 
    sec_full_name_en = Column(CHAR(200)) # 证券英文全称 
    sec_short_name_en = Column(CHAR(200)) # 证券英文简称 
    dyid = Column(CHAR(50)) # 证券展示代码 
    ex_country_cd = Column(CHAR(3)) # 交易市场所属地区 
    trans_curr_cd = Column(CHAR(3)) # 交易货币代码 
    asset_class = Column(CHAR(4)) # 证券类型 
    list_status_cd = Column(CHAR(4)) # 上市状态 
    list_date = Column(DATE) # 上市日期 
    delist_date = Column(DATE) # 摘牌日期 
    party_id = Column(BIGINT) # 机构内部ID 
    dy_use_flg = Column(BIGINT) # DY使用标识 
    update_time = Column(DATETIME) # 更新时间 




from .asset_painter import *
import seaborn as sns

class PortPainter(object):

    @staticmethod
    def plot_mdd(df):
        def rolling_cur_mdd(x):
            if pd.isnull(x).all():
                return np.nan
            x_max = np.fmax.accumulate(x, axis=0)
            return -(1 - np.nanmin(x[-1] / x_max))
        
        mdd_df = df.rolling(window=df.shape[0],min_periods=2).apply(rolling_cur_mdd, raw=True)
        mdd_df.plot(figsize=(16,12),fontsize=20)
        plt.title('策略当前回撤对比',fontsize=25)
        plt.legend(loc='upper left',fontsize=15, bbox_to_anchor=(1,1))
        plt.grid()
        plt.show()
        return mdd_df

    @staticmethod
    def plot_corr(df):
        sns.set(font_scale=2,font='STKaiti')
        plt.figure(figsize=(14,10))
        sns.heatmap(df.pct_change(1).dropna().corr(),cmap='Blues', annot=True)
        plt.show()
        return df




from typing import Optional, Dict
# import datetime
import os
import traceback
import time
import tarfile

from sqlalchemy.orm import sessionmaker
import pandas as pd
import numpy as np

from ...util.mail_retriever import MailAttachmentRetriever, IMAP_SPType, UID_FILE_NAME
from ...util.wechat_bot import WechatBot
from ...util.calculator import Calculator
from ...constant import FOFStrategyType
from ..wrapper.mysql import RawDatabaseConnector, BasicDatabaseConnector, DerivedDatabaseConnector
from ..api.raw import RawDataApi
from ..api.basic import BasicDataApi
from ..api.derived import DerivedDataApi
from ..view.raw_models import HFIndexPrice, HFFundNav, HFFundInfo
from ..view.basic_models import FOFInfo
from ..view.derived_models import FOFNavPublic
from ..fund.raw.other_raw_data_downloader import OtherRawDataDownloader
from ..fund.raw.raw_data_helper import RawDataHelper


class ResearchNAVReader:

    _FUND_INFO_COL_NAME: Dict[str, str] = {
        '策略类型*': 'stg_type',
        '产品简称*': 'desc_name',
        '产品ID*': 'fund_id',
        '成立日期*': 'begin_date',
        '机构ID*': 'agency_id',
        '机构简称': 'agency_name',
        '基金经理': 'manager_name',
        '来源': 'source',
        '净值来源': 'nav_source',
    }

    def __init__(self, read_dir: str, user_name: str, password: str):
        pd.set_option('display.max_rows', None)
        pd.set_option('display.max_columns', None)

        self._read_dir = read_dir
        assert os.path.isdir(self._read_dir), f'arg dump_dir should be a directory (now){self._read_dir}'

        self._user_name = user_name
        self._password = password
        self._wechat_bot = WechatBot()

    @staticmethod
    def _read_for_NAVData(file_path: str) -> Dict[str, int]:
        update_info: Dict[str, int] = {}
        tar = tarfile.open(file_path, 'r:gz', encoding='gb2312')
        for one in tar:
            if not one.name.endswith('.xls') and not one.name.endswith('.xlsx'):
                continue
            if '业绩走势' in one.name:
                continue
            try:
                if one.name.startswith('./'):
                    p_name = one.name[len('./'):].split('.')[0]
                else:
                    p_name = one.name.split('.')[0]
                try:
                    normal_replace_dict = {
                        '・': '·',
                    }
                    for k, v in normal_replace_dict.items():
                        p_name = p_name.replace(k, v)
                    print(f'[_read_for_NAVData] (name in tar){one.name} (p_name){p_name}')
                except UnicodeEncodeError:
                    replace_dict = {
                        '\udc9bK': '汯',
                        '\udcacB': '珺',
                        '\udcd4Z': '訸',
                        '\udcb6G': '禛',
                        '\udc95D': '旸',
                        '\udcf1I': '馡',
                        '\udca9q': '﹒',
                    }
                    for k, v in replace_dict.items():
                        p_name = p_name.replace(k, v)
                    print(f'[_read_for_NAVData] (name in tar (fixed)){p_name}')
                df = pd.read_excel(tar.extractfile(one), na_values=['--'])
                df = df.rename(columns={'日期': 'datetime', '净值(分红再投)': 'nav'})
                df = df[['datetime', 'nav']]
                hf_fund_info = RawDataApi().get_hf_fund_info()
                hf_fund_info = hf_fund_info[hf_fund_info.desc_name == p_name]
                fund_id = hf_fund_info.fund_id.array[0]
                df['fund_id'] = fund_id
                df['datetime'] = pd.to_datetime(df.datetime, infer_datetime_format=True).dt.date
                hf_fund_nav = RawDataApi().get_hf_fund_nav(fund_ids=[fund_id])
                if hf_fund_nav is not None:
                    df_fitted = df.astype(hf_fund_nav.dtypes.to_dict())
                    # merge on all columns
                    df_fitted = df_fitted.merge(hf_fund_nav, how='left', indicator=True, validate='one_to_one')
                    df_fitted = df_fitted[df_fitted._merge == 'left_only'].drop(columns=['_merge'])
                else:
                    df_fitted = df
                print(f'to hf_fund_nav {df_fitted}')
                if not df_fitted.empty:
                    # TODO: 最好原子提交下边两步
                    RawDataApi().delete_hf_fund_nav(fund_id_to_delete=fund_id, datetime_list=df_fitted.datetime.to_list())
                    # 更新到db
                    df_fitted.to_sql(HFFundNav.__table__.name, RawDatabaseConnector().get_engine(), index=False, if_exists='append')

                # 再向 FOFNavPublic 更新一份
                df = df.rename(columns={'fund_id': 'fof_id', 'nav': 'adjusted_nav'})
                df = df[df.adjusted_nav.notna()]
                if df.empty:
                    continue
                df['nav'] = df.adjusted_nav
                df['acc_net_value'] = df.adjusted_nav
                fof_nav_public = DerivedDataApi().get_fof_nav_public(fof_id_list=[fund_id])
                if fof_nav_public is not None:
                    fof_nav_public = fof_nav_public.drop(columns=['volume', 'mv', 'ret', 'create_time', 'update_time', 'is_deleted'])
                    df = df.astype(fof_nav_public.dtypes.to_dict())
                    # merge on all columns
                    df = df.merge(fof_nav_public, how='left', indicator=True, validate='one_to_one')
                    df = df[df._merge == 'left_only'].drop(columns=['_merge'])
                print(f'to fof_nav_public {df}')
                if not df.empty:
                    # TODO: 最好原子提交下边两步
                    DerivedDataApi().delete_fof_nav_public(fof_id_to_delete=fund_id, datetime_list=df.datetime.to_list())
                    # 更新到db
                    df.to_sql(FOFNavPublic.__table__.name, DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')
                    update_info[fund_id] = df.shape[0]

                fof_nav = DerivedDataApi().get_fof_nav_public(fof_id_list=[fund_id])
                if fof_nav is not None and not fof_nav.empty and (fof_nav.shape[0] > 1):
                    fof_nav_public = fof_nav.set_index('datetime')['adjusted_nav']
                    res_status = Calculator.get_stat_result(fof_nav_public.index, fof_nav_public.array)
                    Session = sessionmaker(BasicDatabaseConnector().get_engine())
                    db_session = Session()
                    # '1' 表示公有的 FOFInfo
                    fof_info_to_set = db_session.query(FOFInfo).filter(FOFInfo.manager_id == '1', FOFInfo.fof_id == fund_id).one_or_none()
                    fof_info_to_set.net_asset_value = float(fof_nav.nav.array[-1]) if not pd.isnull(fof_nav.nav.array[-1]) else None
                    fof_info_to_set.acc_unit_value = float(fof_nav.acc_net_value.array[-1]) if not pd.isnull(fof_nav.acc_net_value.array[-1]) else None
                    fof_info_to_set.adjusted_net_value = float(fof_nav_public.array[-1]) if not pd.isnull(fof_nav_public.array[-1]) else None
                    fof_info_to_set.latest_cal_date = res_status.end_date
                    fof_info_to_set.ret_year_to_now = float(res_status.recent_year_ret) if not pd.isnull(res_status.recent_year_ret) else None
                    fof_info_to_set.ret_total = float(res_status.cumu_ret) if not pd.isnull(res_status.cumu_ret) else None
                    fof_info_to_set.ret_ann = float(res_status.annualized_ret) if not pd.isnull(res_status.annualized_ret) else None
                    fof_info_to_set.mdd = float(res_status.mdd) if not pd.isnull(res_status.mdd) else None
                    fof_info_to_set.sharpe = float(res_status.sharpe) if not pd.isnull(res_status.sharpe) else None
                    fof_info_to_set.vol = float(res_status.annualized_vol) if not pd.isnull(res_status.annualized_vol) else None
                    fof_info_to_set.has_nav = True
                    db_session.commit()
                    db_session.close()
            except Exception as e:
                traceback.print_exc()
                print(f'[_read_for_NAVData] failed (err){e}')
        tar.close()
        return update_info

    @staticmethod
    def _read_for_IndexData(file_path: str) -> Dict[str, int]:
        update_info = {}
        tar = tarfile.open(file_path, 'r:gz', encoding='gb2312')
        for one in tar:
            print(f'[_read_for_IndexData] (name in tar){one.name}')
            index_name = one.name.split('/')[-1].split('.')[0]
            if '朝阳永续' in one.name:
                key_name = one.name.split('朝阳永续-')[-1].split('.')[0]
                df = pd.read_excel(tar.extractfile(one), na_values=['--'], thousands=',')
                df = df.rename(columns={'时间': 'index_date', key_name: 'close'})
                df = df[['index_date', 'close']]
                df['calcu_date'] = None
            else:
                if '融智' not in one.name:
                    continue
                    # print(f'[_read_for_IndexData] got special file name {one}')
                key_name = one.name.split('融智-')[-1].split('.')[0]
                # df = pd.read_excel(tar.extractfile(one), header=3, na_values=['--'])
                # df = df.rename(columns={'指数日期': 'index_date', '指数计算日期': 'calcu_date', key_name: 'close'})
                # df = df[['index_date', 'calcu_date', 'close']]
                # df['calcu_date'] = pd.to_datetime(df.calcu_date, infer_datetime_format=True).dt.date
                df = pd.read_excel(tar.extractfile(one), header=0, na_values=['--'])
                df = df.rename(columns={'点位日期': 'index_date', '指数点位': 'close'})
                df = df[['index_date', 'close']]
                df['calcu_date'] = None
            df = df[df.close.notna()]
            asset_info = BasicDataApi().get_asset_info_by_type(['策略指数'])
            asset_info = asset_info[asset_info.real_name == index_name]
            if asset_info.empty:
                assert False, f'[_read_for_IndexData] can not find name of index {index_name}'
            index_id = asset_info.real_id.array[0]
            df['index_id'] = index_id
            df['index_date'] = pd.to_datetime(df.index_date, infer_datetime_format=True).dt.date
            hf_index_price = RawDataApi().get_hf_index_price(index_ids=[index_id])
            if hf_index_price is not None:
                df = df.astype(hf_index_price.dtypes.to_dict())
                # merge on all columns
                df = df.merge(hf_index_price, how='left', indicator=True, validate='one_to_one')
                df = df[df._merge == 'left_only'].drop(columns=['_merge'])
            print(df)
            if not df.empty:
                # TODO: 最好原子提交下边两步
                RawDataApi().delete_hf_index_price(index_id, df.index_date.to_list())
                # 更新到db
                df.to_sql(HFIndexPrice.__table__.name, RawDatabaseConnector().get_engine(), index=False, if_exists='append')
                update_info[key_name] = df.shape[0]
        tar.close()
        return update_info

    @staticmethod
    def _read_for_FundInfo(file_path: str):
        tar = tarfile.open(file_path, 'r:gz', encoding='gb2312')
        for one in tar:
            df = pd.read_excel(tar.extractfile(one), sheet_name='产品列表', header=1, na_values=['--'])
            df = df.loc[:, df.notna().any(axis=0)].dropna(subset=['产品简称*']).drop(columns=['标签'])
            df = df[list(ResearchNAVReader._FUND_INFO_COL_NAME.keys())].rename(columns=ResearchNAVReader._FUND_INFO_COL_NAME)
            df['begin_date'] = df.begin_date.dt.date
            df = df.drop_duplicates()
            hf_fund_info = RawDataApi().get_hf_fund_info()
            if hf_fund_info is not None:
                to_inserted_df = df.astype(hf_fund_info.dtypes.to_dict())
                # merge on all columns
                to_inserted_df = to_inserted_df.merge(hf_fund_info, how='left', indicator=True, validate='one_to_one')
                to_inserted_df = to_inserted_df[to_inserted_df._merge == 'left_only'].drop(columns=['_merge'])
            else:
                to_inserted_df = df
            print(to_inserted_df)
            if not to_inserted_df.empty:
                to_inserted_df = to_inserted_df.replace({np.nan: None}).drop_duplicates(subset=['fund_id'])
                # TODO: 最好原子提交下边两步
                RawDataApi().delete_hf_fund_info(fund_id_list=to_inserted_df.fund_id.to_list())
                # 更新到db
                to_inserted_df.to_sql(HFFundInfo.__table__.name, RawDatabaseConnector().get_engine(), index=False, if_exists='append')

            existed_fof_info = BasicDataApi().get_fof_info('1', list(df.fund_id.unique()))
            lacked_fof_ids = set(df.fund_id.array) - set(existed_fof_info.fof_id.array)
            if lacked_fof_ids:
                df = df[df.fund_id.isin(lacked_fof_ids)].rename(columns={
                    'fund_id': 'fof_id',
                    'begin_date': 'established_date',
                    'agency_id': 'management_ids',
                    'agency_name': 'admin',
                    'manager_name': 'fof_manager',
                })
                df['strategy_type'] = df.stg_type.map(FOFStrategyType.get_type_of_names())
                df['fof_name'] = df['desc_name']
                df['manager_id'] = '1'
                df = df.drop(columns=['source', 'nav_source', 'stg_type'])
                df = df.drop_duplicates(subset='fof_id', keep='first')
                print(df)
                df.to_sql(FOFInfo.__table__.name, BasicDatabaseConnector().get_engine(), index=False, if_exists='append')

    def _notify_error_event(self, err_msg: str):
        print(f'[read_navs_and_dump_to_db] {err_msg}')
        # self._wechat_bot.send_hedge_fund_nav_update_failed(err_msg)

    def read_navs_and_dump_to_db(self):
        try:
            with open(os.path.join(self._read_dir, UID_FILE_NAME), 'rb') as f:
                uid_last = f.read()
                if not uid_last:
                    uid_last = None
        except Exception as e:
            self._notify_error_event(f'read uid file failed (e){e}, use None instead(read all emails)')
            uid_last = None

        try:
            mar = MailAttachmentRetriever(self._read_dir, ['tar.gz', ])
            data = mar.get_excels(IMAP_SPType.IMAP_QQ, self._user_name, self._password, uid_last)
        except Exception as e:
            self._notify_error_event(f'FATAL ERROR!! get new data of hedge fund nav failed (e){e}')
            return

        rdh = RawDataHelper()
        uid_last_succeed: Optional[bytes] = None
        for name, comp_date in data.items():
            uid, file_path = comp_date
            if not name.endswith('.tar.gz'):
                continue

            try:
                if '净值数据' in name:
                    update_info = ResearchNAVReader._read_for_NAVData(file_path)

                    ordd = OtherRawDataDownloader(rdh)
                    ret = ordd.hf_fund_nav_miss()
                    assert not ret, 'update table of hf fund nav miss failed!!'

                    # self._wechat_bot.send_research_data_update('净值', update_info)
                elif '指数数据' in name:
                    update_info = ResearchNAVReader._read_for_IndexData(file_path)

                    # self._wechat_bot.send_research_data_update('指数', update_info)
                elif '跟踪产品' in name:
                    ResearchNAVReader._read_for_FundInfo(file_path)

                    ordd = OtherRawDataDownloader(rdh)
                    ret = ordd.hf_fund_nav_miss()
                    assert not ret, 'update table of hf fund nav miss failed!!'

                    # self._wechat_bot.send_research_data_update('基金信息', update_info)
                else:
                    raise NotImplementedError('unknown research nav file from attachment')
            except Exception as e:
                traceback.print_exc()
                self._notify_error_event(f'{e} (parse) (name){name} (file_path){file_path}')
                continue

            uid_last_succeed = uid
            time.sleep(1)

        # 记录下成功的最后一个uid
        if uid_last_succeed is not None:
            with open(os.path.join(self._read_dir, UID_FILE_NAME), 'wb') as f:
                f.write(uid_last_succeed)
        return


if __name__ == '__main__':
    try:
        email_data_dir = os.environ['SURFING_EMAIL_DATA_DIR']
        user_name = os.environ['SURFING_EMAIL_USER_NAME']
        password = os.environ['SURFING_EMAIL_PASSWORD']
    except KeyError as e:
        import sys
        sys.exit(f'can not found enough params in env (e){e}')

    r_nav_r = ResearchNAVReader(email_data_dir, user_name, password)
    r_nav_r.read_navs_and_dump_to_db()



import re

class KeywordParser(object):

    def parse(self, word_tuples, text, result):
        words = [x[0] for x in word_tuples]

        if '我的持仓' in text:
            result['type'] = 'cczx'
            result['data']['target'] = 'all'
        elif '我的净值' in text:
            result['type'] = 'cczx'
            result['data']['target'] = 'nav'
        elif '交易记录' in text:
            result['type'] = 'cczx'
            result['data']['target'] = 'trades'
        # # 处理通用咨询 (tyzx)
        # if '投资' in text and ('策略' in text or ('建立' in text and '组合' in text)):
        #     result['type'] = 'tyzx'
        #     result['data'] = {
        #         'target': '投资策略'
        #     }
        # elif '收益' in text and '风险' in text and '预估' in text:
        #     result['type'] = 'tyzx'
        #     result['target'] = '投资策略'
        # elif '持有' in text and ('多久' in text or '时间' in text):
        #     result['type'] = 'tyzx'
        #     result['target'] = '持有时间'
        # elif '赎回' in text:
        #     result['type'] = 'tyzx'
        #     result['target'] = '持有时间'
        # elif '费用' in text or '收费' in text or '交易费' in text:
        #     result['type'] = 'tyzx'
        #     result['target'] = '收费方式'
        # elif '交易' in text:
        #     if '记录' in text:
        #         result['type'] = 'cczx'
        #         result['target'] = 'trading_history'
        #     elif '如何' in text:
        #         result['type'] = 'tyzx'
        #         result['target'] = '收费方式'
        # elif '年化' in text and '收益' in text:
        #     arr = re.findall(r"\d+\.?\d*", text)
        #     if len(arr) > 0:
        #         ret = float(arr[0])
        #         if ret > 1:
        #             ret = ret / 100
        #         result['type'] = 'zhtj'
        #         result['target'] = 'ret'
        #         result['ret'] = ret
        # elif '高' in text and '收益' in text:
        #     result['type'] = 'zhtj'
        #     result['target'] = 'ret'
        #     result['ret'] = 'high'
        # elif '回撤' in text:
        #     arr = re.findall(r"\d+\.?\d*", text)
        #     if len(arr) > 0:
        #         mdd = float(arr[0])
        #         if mdd > 1:
        #             mdd = mdd / 100
        #         result['type'] = 'zhtj'
        #         result['target'] = 'mdd'
        #         result['mdd'] = mdd
        # elif '低' in text and '风险' in text:
        #     result['type'] = 'zhtj'
        #     result['target'] = 'mdd'
        #     result['mdd'] = 'low'
        # elif '股票' in text:
        #     if '为何' in text or '为什么' in text:
        #         if '收益' in text and ('不明显' in text or '下降' in text):
        #             result['type'] = 'yjfx'
        #             result['baseline'] = 'stock'
        #         else:
        #             result['type'] = 'chfx'
        #             result['target'] = 'asset'
        #             result['asset'] = 'stock'
        #     else:
        #         result['type'] = 'hqzx'
        #         result['target'] = 'asset'
        #         result['asset'] = 'stock'
        # elif '信用债' in text:
        #     if '为何' in text or '为什么' in text:
        #         result['type'] = 'chfx'
        #         result['target'] = 'asset'
        #         result['asset'] = 'credit_debt'
        #     else:
        #         result['type'] = 'hqzx'
        #         result['target'] = 'asset'
        #         result['asset'] = 'credit_debt'
        # elif 'target' in result.keys() and result['target'] == 'fund':
        #     if '为何' in text or '为什么' in text:
        #         result['type'] = 'chfx'
        #     else:
        #         result['type'] = 'hqzx'
        # elif '组合' and '收益' in text:
        #     result['type'] = 'cczx'
        #     result['target'] = 'nav'
        # elif '组合' in text and '板' in text:
        #     result['type'] = 'cczx'
        #     result['target'] = 'industry'
        # elif '大类' in text and ('配比' in text or '表现' in text):
        #     if '为何' in text or '为什么' in text:
        #         result['type'] = 'chfx'
        #         result['target'] = 'asset'
        #         result['asset'] = 'all'
        #     else:
        #         result['type'] = 'cczx'
        #         result['target'] = 'asset'
        # elif '持仓' in text:
        #     result['type'] = 'cczx'
        #     result['target'] = 'fund'
        # elif '基金' in text:
        #     if '怎么' in text or '为什么' in text or '为何' in text:
        #         result['type'] = 'chfx'
        #         result['target'] = 'fund'
        #         result['fund'] = 'all'
        #     else:
        #         result['type'] = 'cczx'
        #         result['target'] = 'fund'
        # elif '业绩' in text:
        #     result['type'] = 'cczx'
        #     result['target'] = 'backtest'
        # elif '国债' in text:
        #     result['type'] = 'chfx'
        #     result['target'] = 'asset'
        #     result['asset'] = 'national_debt'
        # elif '净值' in text and '下降' in text:
        #     result['type'] = 'yjfx'
        # elif '组合' in text and '亏损' in text:
        #     result['type'] = 'yjfx'

        return result

        
    
    
    
        